import inspect
import pathlib
import uuid
from itertools import permutations
from typing import List

import pandas as pd
import pytest

import hamilton.graph_utils
import hamilton.htypes
from hamilton import ad_hoc_utils, base, graph, node
from hamilton.execution import graph_functions
from hamilton.function_modifiers import schema
from hamilton.lifecycle import base as lifecycle_base
from hamilton.node import NodeType

import tests.resources.bad_functions
import tests.resources.compatible_input_types
import tests.resources.config_modifier
import tests.resources.cyclic_functions
import tests.resources.dummy_functions
import tests.resources.dummy_functions_module_override
import tests.resources.extract_column_nodes
import tests.resources.extract_columns_execution_count
import tests.resources.functions_with_generics
import tests.resources.incompatible_input_types
import tests.resources.layered_decorators
import tests.resources.multiple_decorators_together
import tests.resources.optional_dependencies
import tests.resources.parametrized_inputs
import tests.resources.parametrized_nodes
import tests.resources.test_default_args
import tests.resources.typing_vs_not_typing


def test_find_functions():
    """Tests that we filter out _ functions when passed a module and don't pull in anything from the imports."""
    expected = [
        ("A", tests.resources.dummy_functions.A),
        ("B", tests.resources.dummy_functions.B),
        ("C", tests.resources.dummy_functions.C),
    ]
    actual = hamilton.graph_utils.find_functions(tests.resources.dummy_functions)
    assert len(actual) == len(expected)
    assert actual == expected


def test_find_functions_from_temporary_function_module():
    """Tests that we handle the TemporaryFunctionModule object correctly."""
    expected = [
        ("A", tests.resources.dummy_functions.A),
        ("B", tests.resources.dummy_functions.B),
        ("C", tests.resources.dummy_functions.C),
    ]
    func_module = ad_hoc_utils.create_temporary_module(
        tests.resources.dummy_functions.A,
        tests.resources.dummy_functions.B,
        tests.resources.dummy_functions.C,
    )
    actual = hamilton.graph_utils.find_functions(func_module)
    assert len(actual) == len(expected)
    assert [node_name for node_name, _ in actual] == [node_name for node_name, _ in expected]
    assert [fn.__code__ for _, fn in actual] == [
        fn.__code__ for _, fn in expected
    ]  # easy way to say they're the same


def test_add_dependency_missing_param_type():
    """Tests case that we error if types are missing from a parameter."""
    with pytest.raises(ValueError):
        a_sig = inspect.signature(tests.resources.bad_functions.A)
        node.Node(
            "A", a_sig.return_annotation, "A doc", tests.resources.bad_functions.A
        )  # should error out


def test_add_dependency_missing_function_type():
    """Tests case that we error if types are missing from a function."""
    with pytest.raises(ValueError):
        b_sig = inspect.signature(tests.resources.bad_functions.B)
        node.Node(
            "B", b_sig.return_annotation, "B doc", tests.resources.bad_functions.B
        )  # should error out


def test_add_dependency_strict_node_dependencies():
    """Tests that we add node dependencies between functions correctly.

    Setup here is: B depends on A. So A is depended on by B. B is not depended on by anyone.
    """
    b_sig = inspect.signature(tests.resources.dummy_functions.B)
    func_node = node.Node("B", b_sig.return_annotation, "B doc", tests.resources.dummy_functions.B)
    func_name = "B"
    nodes = {
        "A": node.Node(
            "A",
            inspect.signature(tests.resources.dummy_functions.A).return_annotation,
            "A doc",
            tests.resources.dummy_functions.A,
        )
    }
    param_name = "A"
    param_type = b_sig.parameters["A"].annotation

    graph.add_dependency(
        func_node,
        func_name,
        nodes,
        param_name,
        param_type,
        lifecycle_base.LifecycleAdapterSet(),
    )
    assert nodes["A"] == func_node.dependencies[0]
    assert func_node.depended_on_by == []


def test_add_dependency_input_nodes_mismatch_on_types():
    """Tests that if two functions request an input that has incompatible types, we error out."""
    b_sig = inspect.signature(tests.resources.incompatible_input_types.b)
    c_sig = inspect.signature(tests.resources.incompatible_input_types.c)

    nodes = {
        "b": node.Node.from_fn(tests.resources.incompatible_input_types.b),
        "c": node.Node.from_fn(tests.resources.incompatible_input_types.c),
    }
    nodes["b"]._originating_functions = (tests.resources.incompatible_input_types.b,)
    nodes["c"]._originating_functions = (tests.resources.incompatible_input_types.c,)
    param_name = "a"

    # this adds 'a' to nodes
    graph.add_dependency(
        nodes["b"],
        "b",
        nodes,
        param_name,
        b_sig.parameters[param_name].annotation,
        lifecycle_base.LifecycleAdapterSet(),
    )

    assert "a" in nodes

    # adding dependency of c on a should fail because the types are incompatible
    with pytest.raises(ValueError):
        graph.add_dependency(
            nodes["c"],
            "c",
            nodes,
            param_name,
            c_sig.parameters[param_name].annotation,
            lifecycle_base.LifecycleAdapterSet(),
        )


def test_add_dependency_input_nodes_mismatch_on_types_complex():
    """Tests a more complex scenario we don't support right now with input types."""
    e_sig = inspect.signature(tests.resources.incompatible_input_types.e)
    f_sig = inspect.signature(tests.resources.incompatible_input_types.f)

    nodes = {
        "e": node.Node.from_fn(tests.resources.incompatible_input_types.e),
        "f": node.Node.from_fn(tests.resources.incompatible_input_types.f),
    }
    nodes["e"]._originating_functions = (tests.resources.incompatible_input_types.e,)
    nodes["f"]._originating_functions = (tests.resources.incompatible_input_types.f,)
    param_name = "d"

    # this adds 'a' to nodes
    graph.add_dependency(
        nodes["e"],
        "e",
        nodes,
        param_name,
        e_sig.parameters[param_name].annotation,
        lifecycle_base.LifecycleAdapterSet(),
    )

    assert "d" in nodes

    # adding dependency of c on a should fail because the types are incompatible
    with pytest.raises(ValueError):
        graph.add_dependency(
            nodes["e"],
            "e",
            nodes,
            param_name,
            f_sig.parameters[param_name].annotation,
            lifecycle_base.LifecycleAdapterSet(),
        )


def test_add_dependency_input_nodes_compatible_types():
    """Tests that if functions request an input that we correctly accept compatible types."""
    b_sig = inspect.signature(tests.resources.compatible_input_types.b)
    c_sig = inspect.signature(tests.resources.compatible_input_types.c)
    d_sig = inspect.signature(tests.resources.compatible_input_types.d)

    nodes = {
        "b": node.Node.from_fn(tests.resources.compatible_input_types.b),
        "c": node.Node.from_fn(tests.resources.compatible_input_types.c),
        "d": node.Node.from_fn(tests.resources.compatible_input_types.d),
    }
    nodes["b"]._originating_functions = (tests.resources.compatible_input_types.b,)
    nodes["c"]._originating_functions = (tests.resources.compatible_input_types.c,)
    nodes["d"]._originating_functions = (tests.resources.compatible_input_types.d,)
    # what we want to add
    param_name = "a"

    # this adds 'a' to nodes
    graph.add_dependency(
        nodes["b"],
        "b",
        nodes,
        param_name,
        b_sig.parameters[param_name].annotation,
        lifecycle_base.LifecycleAdapterSet(),
    )

    assert "a" in nodes

    # this adds 'a' to 'c' as well.
    graph.add_dependency(
        nodes["c"],
        "c",
        nodes,
        param_name,
        c_sig.parameters[param_name].annotation,
        lifecycle_base.LifecycleAdapterSet(),
    )

    # test that we shrink the type to the tighter type
    assert nodes["a"].type == str

    graph.add_dependency(
        nodes["d"],
        "d",
        nodes,
        param_name,
        d_sig.parameters[param_name].annotation,
        lifecycle_base.LifecycleAdapterSet(),
    )


def test_add_dependency_input_nodes_compatible_types_order_check():
    """Tests that if functions request an input that we correctly accept compatible types independent of order.

    This just reorders test_add_dependency_input_nodes_compatible_types to ensure the outcome does not change.
    """
    b_sig = inspect.signature(tests.resources.compatible_input_types.b)
    c_sig = inspect.signature(tests.resources.compatible_input_types.c)
    d_sig = inspect.signature(tests.resources.compatible_input_types.d)

    nodes = {
        "b": node.Node.from_fn(tests.resources.compatible_input_types.b),
        "c": node.Node.from_fn(tests.resources.compatible_input_types.c),
        "d": node.Node.from_fn(tests.resources.compatible_input_types.d),
    }
    nodes["b"]._originating_functions = (tests.resources.compatible_input_types.b,)
    nodes["c"]._originating_functions = (tests.resources.compatible_input_types.c,)
    nodes["d"]._originating_functions = (tests.resources.compatible_input_types.d,)
    # what we want to add
    param_name = "a"

    # this adds 'a' to nodes
    graph.add_dependency(
        nodes["c"],
        "c",
        nodes,
        param_name,
        c_sig.parameters[param_name].annotation,
        lifecycle_base.LifecycleAdapterSet(),
    )

    assert "a" in nodes
    assert nodes["a"].type == str

    # this adds 'a' to 'c' as well.
    graph.add_dependency(
        nodes["b"],
        "b",
        nodes,
        param_name,
        b_sig.parameters[param_name].annotation,
        lifecycle_base.LifecycleAdapterSet(),
    )

    # test that type didn't change
    assert nodes["a"].type == str

    graph.add_dependency(
        nodes["d"],
        "d",
        nodes,
        param_name,
        d_sig.parameters[param_name].annotation,
        lifecycle_base.LifecycleAdapterSet(),
    )


def test_typing_to_primitive_conversion():
    """Tests that we can mix function output being typing type, and dependent function using primitive type."""
    b_sig = inspect.signature(tests.resources.typing_vs_not_typing.B)
    func_node = node.Node(
        "B", b_sig.return_annotation, "B doc", tests.resources.typing_vs_not_typing.B
    )
    func_name = "B"
    nodes = {
        "A": node.Node(
            "A",
            inspect.signature(tests.resources.typing_vs_not_typing.A).return_annotation,
            "A doc",
            tests.resources.typing_vs_not_typing.A,
        )
    }
    param_name = "A"
    param_type = b_sig.parameters["A"].annotation

    graph.add_dependency(
        func_node,
        func_name,
        nodes,
        param_name,
        param_type,
        lifecycle_base.LifecycleAdapterSet(),
    )
    assert nodes["A"] == func_node.dependencies[0]
    assert func_node.depended_on_by == []


def test_primitive_to_typing_conversion():
    """Tests that we can mix function output being a primitive type, and dependent function using typing type."""
    b_sig = inspect.signature(tests.resources.typing_vs_not_typing.B2)
    func_node = node.Node(
        "B2", b_sig.return_annotation, "B2 doc", tests.resources.typing_vs_not_typing.B2
    )
    func_name = "B2"
    nodes = {
        "A2": node.Node(
            "A2",
            inspect.signature(tests.resources.typing_vs_not_typing.A2).return_annotation,
            "A2 doc",
            tests.resources.typing_vs_not_typing.A2,
        )
    }
    param_name = "A2"
    param_type = b_sig.parameters["A2"].annotation

    graph.add_dependency(
        func_node,
        func_name,
        nodes,
        param_name,
        param_type,
        lifecycle_base.LifecycleAdapterSet(),
    )
    assert nodes["A2"] == func_node.dependencies[0]
    assert func_node.depended_on_by == []


def test_throwing_error_on_incompatible_types():
    """Tests we error on incompatible types."""
    d_sig = inspect.signature(tests.resources.bad_functions.D)
    func_node = node.Node("D", d_sig.return_annotation, "D doc", tests.resources.bad_functions.D)
    func_name = "D"
    nodes = {
        "C": node.Node(
            "C",
            inspect.signature(tests.resources.bad_functions.C).return_annotation,
            "C doc",
            tests.resources.bad_functions.C,
        )
    }
    param_name = "C"
    param_type = d_sig.parameters["C"].annotation
    with pytest.raises(ValueError):
        graph.add_dependency(
            func_node,
            func_name,
            nodes,
            param_name,
            param_type,
            lifecycle_base.LifecycleAdapterSet(),
        )


def test_add_dependency_user_nodes():
    """Tests that we add node user defined dependencies correctly.

    Setup here is: A depends on b and c. But we're only doing one call. So expecting A having 'b' as a dependency,
    and 'b' is depended on by A.
    """
    a_sig = inspect.signature(tests.resources.dummy_functions.A)
    func_node = node.Node("A", a_sig.return_annotation, "A doc", tests.resources.dummy_functions.A)
    func_name = "A"
    nodes = {}
    param_name = "b"
    param_type = a_sig.parameters["b"].annotation

    graph.add_dependency(
        func_node,
        func_name,
        nodes,
        param_name,
        param_type,
        lifecycle_base.LifecycleAdapterSet(),
    )
    # user node is created and added to nodes.
    assert nodes["b"] == func_node.dependencies[0]
    assert nodes["b"].depended_on_by[0] == func_node
    assert func_node.depended_on_by == []


def create_testing_nodes():
    """Helper function for creating the nodes represented in dummy_functions.py."""
    nodes = {
        "A": node.Node.from_fn(fn=tests.resources.dummy_functions.A, name="A"),
        "B": node.Node.from_fn(fn=tests.resources.dummy_functions.B, name="B"),
        "C": node.Node.from_fn(fn=tests.resources.dummy_functions.C, name="C"),
        "b": node.Node(
            "b",
            inspect.signature(tests.resources.dummy_functions.A).parameters["b"].annotation,
            node_source=NodeType.EXTERNAL,
        ),
        "c": node.Node(
            "c",
            inspect.signature(tests.resources.dummy_functions.A).parameters["c"].annotation,
            node_source=NodeType.EXTERNAL,
        ),
    }
    nodes["A"].dependencies.append(nodes["b"])
    nodes["A"].dependencies.append(nodes["c"])
    nodes["A"].depended_on_by.append(nodes["B"])
    nodes["A"].depended_on_by.append(nodes["C"])
    nodes["b"].depended_on_by.append(nodes["A"])
    nodes["c"].depended_on_by.append(nodes["A"])
    nodes["B"].dependencies.append(nodes["A"])
    nodes["C"].dependencies.append(nodes["A"])
    return nodes


def create_testing_nodes_override_B():
    """Helper function for creating the nodes represented in dummy_functions.py
    with node B overridden by dummy_functions_module_override.py."""
    nodes = {
        "A": node.Node.from_fn(fn=tests.resources.dummy_functions.A, name="A"),
        "B": node.Node.from_fn(fn=tests.resources.dummy_functions_module_override.B, name="B"),
        "C": node.Node.from_fn(fn=tests.resources.dummy_functions.C, name="C"),
        "b": node.Node(
            "b",
            inspect.signature(tests.resources.dummy_functions.A).parameters["b"].annotation,
            node_source=NodeType.EXTERNAL,
        ),
        "c": node.Node(
            "c",
            inspect.signature(tests.resources.dummy_functions.A).parameters["c"].annotation,
            node_source=NodeType.EXTERNAL,
        ),
    }
    nodes["A"].dependencies.append(nodes["b"])
    nodes["A"].dependencies.append(nodes["c"])
    nodes["A"].depended_on_by.append(nodes["B"])
    nodes["A"].depended_on_by.append(nodes["C"])
    nodes["b"].depended_on_by.append(nodes["A"])
    nodes["c"].depended_on_by.append(nodes["A"])
    nodes["B"].dependencies.append(nodes["A"])
    nodes["C"].dependencies.append(nodes["A"])
    return nodes


def test_create_function_graph_simple():
    """Tests that we create a simple function graph."""
    expected = create_testing_nodes()
    actual = graph.create_function_graph(tests.resources.dummy_functions, config={})
    assert actual == expected


def test_create_function_graph_with_override():
    """Tests that we can override nodes from later modules in function graph."""
    override_expected = create_testing_nodes_override_B()
    override_actual = graph.create_function_graph(
        tests.resources.dummy_functions,
        tests.resources.dummy_functions_module_override,
        config={},
        allow_module_overrides=True,
    )
    assert override_expected == override_actual


def test_execute():
    """Tests graph execution along with basic memoization since A is depended on by two functions."""
    nodes = create_testing_nodes()
    inputs = {"b": 2, "c": 5}
    expected = {"A": 7, "B": 49, "C": 14, "b": 2, "c": 5}
    actual = graph_functions.execute_subdag(nodes=nodes.values(), inputs=inputs)
    assert actual == expected
    actual = graph_functions.execute_subdag(nodes=nodes.values(), inputs=inputs, overrides={"A": 8})
    assert actual["A"] == 8


def test_get_required_functions():
    """Exercises getting the subset of the graph for computation on the toy example we have constructed."""
    nodes = create_testing_nodes()
    final_vars = ["A", "B"]
    expected_user_nodes = {nodes["b"], nodes["c"]}
    expected_nodes = {nodes["A"], nodes["B"], nodes["b"], nodes["c"]}  # we skip 'C'
    fg = graph.FunctionGraph.from_modules(tests.resources.dummy_functions, config={})
    actual_nodes, actual_ud_nodes = fg.get_upstream_nodes(final_vars)
    assert actual_nodes == expected_nodes
    assert actual_ud_nodes == expected_user_nodes


def test_get_downstream_nodes():
    """Exercises getting the downstream subset of the graph for computation on the toy example we have constructed."""
    nodes = create_testing_nodes()
    var_changes = ["A"]
    expected_nodes = {nodes["B"], nodes["C"], nodes["A"]}
    # expected_nodes = {nodes['A'], nodes['B'], nodes['b'], nodes['c']}  # we skip 'C'
    fg = graph.FunctionGraph.from_modules(tests.resources.dummy_functions, config={})
    actual_nodes = fg.get_downstream_nodes(var_changes)
    assert actual_nodes == expected_nodes


def test_function_graph_from_multiple_sources():
    fg = graph.FunctionGraph.from_modules(
        tests.resources.dummy_functions, tests.resources.parametrized_nodes, config={}
    )
    assert len(fg.get_nodes()) == 8  # we take the union of all of them, and want to test that


def test_end_to_end_with_parametrized_nodes():
    """Tests that a simple function graph with parametrized nodes works end-to-end"""
    fg = graph.FunctionGraph.from_modules(tests.resources.parametrized_nodes, config={})
    results = fg.execute(fg.get_nodes(), {})
    assert results == {"parametrized_1": 1, "parametrized_2": 2, "parametrized_3": 3}


def test_end_to_end_with_parametrized_inputs():
    fg = graph.FunctionGraph.from_modules(
        tests.resources.parametrized_inputs, config={"static_value": 3}
    )
    results = fg.execute(fg.get_nodes())
    assert results == {
        "input_1": 1,
        "input_2": 2,
        "input_3": 3,
        "output_1": 1 + 3,
        "output_2": 2 + 3,
        "output_12": 1 + 2 + 3,
        "output_123": 1 + 2 + 3 + 3,
        "static_value": 3,
    }


def test_get_required_functions_askfor_config():
    """Tests that a simple function graph with parametrized nodes works end-to-end"""
    fg = graph.FunctionGraph.from_modules(tests.resources.parametrized_nodes, config={"a": 1})
    nodes, user_nodes = fg.get_upstream_nodes(["a", "parametrized_1"])
    (n,) = user_nodes
    assert n.name == "a"
    results = fg.execute(user_nodes)
    assert results == {"a": 1}


def test_end_to_end_with_column_extractor_nodes():
    """Tests that a simple function graph with nodes that extract columns works end-to-end"""
    fg = graph.FunctionGraph.from_modules(tests.resources.extract_column_nodes, config={})
    nodes = fg.get_nodes()
    results = fg.execute(nodes, {}, {})
    df_expected = tests.resources.extract_column_nodes.generate_df()
    pd.testing.assert_series_equal(results["col_1"], df_expected["col_1"])
    pd.testing.assert_series_equal(results["col_2"], df_expected["col_2"])
    pd.testing.assert_frame_equal(results["generate_df"], df_expected)
    assert (
        nodes[0].documentation == "Function that should be parametrized to form multiple functions"
    )


def test_end_to_end_with_multiple_decorators():
    """Tests that a simple function graph with multiple decorators on a function works end-to-end"""
    fg = graph.FunctionGraph.from_modules(
        tests.resources.multiple_decorators_together,
        config={"param0": 3, "param1": 1, "in_value1": 42, "in_value2": "string_value"},
    )
    nodes = fg.get_nodes()
    # To help debug issues:
    # nodez, user_nodes = fg.get_upstream_nodes([n.name for n in nodes],
    #                                           {"param0": 3, "param1": 1,
    #                                            "in_value1": 42, "in_value2": "string_value"})
    # fg.display(
    #     nodez,
    #     user_nodes,
    #     "all_multiple_decorators",
    #     render_kwargs=None,
    #     graphviz_kwargs=None,
    # )
    results = fg.execute(nodes, {}, {})
    df_expected = tests.resources.multiple_decorators_together._sum_multiply(3, 1, 2)
    dict_expected = tests.resources.multiple_decorators_together._sum(3, 1, 2)
    pd.testing.assert_series_equal(results["param1b"], df_expected["param1b"])
    pd.testing.assert_frame_equal(results["to_modify"], df_expected)
    assert results["total"] == dict_expected["total"]
    assert results["to_modify_2"] == dict_expected
    node_dict = {n.name: n for n in nodes}
    print(sorted(list(node_dict.keys())))
    assert (
        node_dict["to_modify"].documentation
        == "This is a dummy function showing extract_columns with does."
    )
    assert (
        node_dict["to_modify_2"].documentation
        == "This is a dummy function showing extract_fields with does."
    )
    # tag only applies right now to outer most node layer
    assert node_dict["uber_decorated_function"].tags == {
        "module": "tests.resources.multiple_decorators_together"
    }  # tags are not propagated
    assert node_dict["out_value1"].tags == {
        "module": "tests.resources.multiple_decorators_together",
        "test_key": "test-value",
    }
    assert node_dict["out_value2"].tags == {
        "module": "tests.resources.multiple_decorators_together",
        "test_key": "test-value",
    }


def test_end_to_end_with_config_modifier():
    config = {
        "fn_1_version": 1,
    }
    fg = graph.FunctionGraph.from_modules(tests.resources.config_modifier, config=config)
    results = fg.execute(fg.get_nodes(), {}, {})
    assert results["fn"] == "version_1"

    config = {
        "fn_1_version": 2,
    }
    fg = graph.FunctionGraph.from_modules(tests.resources.config_modifier, config=config)
    results = fg.execute(fg.get_nodes(), {}, {})
    assert results["fn"] == "version_2"
    config = {
        "fn_1_version": 3,
    }
    fg = graph.FunctionGraph.from_modules(tests.resources.config_modifier, config=config)
    results = fg.execute(fg.get_nodes(), {}, {})
    assert results["fn"] == "version_3"


def test_non_required_nodes():
    fg = graph.FunctionGraph.from_modules(
        tests.resources.test_default_args, config={"required": 10}
    )
    results = fg.execute(
        # D is not on the execution path, so it should not break things
        [n for n in fg.get_nodes() if n.node_role == NodeType.STANDARD and n.name != "D"],
        {},
        {},
    )
    assert results["A"] == 10
    fg = graph.FunctionGraph.from_modules(
        tests.resources.test_default_args, config={"required": 10, "defaults_to_zero": 1}
    )
    results = fg.execute(
        [n for n in fg.get_nodes() if n.node_role == NodeType.STANDARD],
        {},
        {},
    )
    assert results["A"] == 11
    assert results["D"] == 2


def test_config_can_override():
    config = {"new_param": "new_value"}
    fg = graph.FunctionGraph.from_modules(tests.resources.config_modifier, config=config)
    out = fg.execute([n for n in fg.get_nodes()])
    assert out["new_param"] == "new_value"


def test_function_graph_has_cycles_true():
    """Tests whether we catch a graph with cycles -- and expected behaviors"""
    fg = graph.FunctionGraph.from_modules(tests.resources.cyclic_functions, config={"b": 2, "c": 1})
    all_nodes = fg.get_nodes()
    nodes = [n for n in all_nodes if not n.user_defined]
    user_nodes = [n for n in all_nodes if n.user_defined]
    assert fg.has_cycles(nodes, user_nodes) is True
    required_nodes, required_user_nodes = fg.get_upstream_nodes(["A", "B", "C"])
    assert required_nodes == set(nodes + user_nodes)
    assert required_user_nodes == set(user_nodes)
    # We don't want to support this behavior officially -- but this works:
    # result = fg.execute([n for n in nodes if n.name == 'B'], overrides={'A': 1, 'D': 2})
    # assert len(result) == 3
    # assert result['B'] == 3
    with pytest.raises(
        RecursionError
    ):  # throw recursion error when we don't have a way to short circuit
        fg.execute([n for n in nodes if n.name == "B"])


def test_function_graph_has_cycles_false():
    """Tests whether we catch a graph with cycles"""
    fg = graph.FunctionGraph.from_modules(tests.resources.dummy_functions, config={"b": 1, "c": 2})
    all_nodes = fg.get_nodes()
    # checks it two ways
    nodes = [n for n in all_nodes if not n.user_defined]
    user_nodes = [n for n in all_nodes if n.user_defined]
    assert fg.has_cycles(nodes, user_nodes) is False
    # this is called by the driver
    nodes, user_nodes = fg.get_upstream_nodes(["A", "B", "C"])
    assert fg.has_cycles(nodes, user_nodes) is False


def test_function_graph_display_content(tmp_path: pathlib.Path):
    """Tests that display saves a file"""
    dot_file_path = tmp_path / "dag"
    config = {"b": 1, "c": 2}
    fg = graph.FunctionGraph.from_modules(tests.resources.dummy_functions, config=config)
    node_modifiers = {"B": {graph.VisualizationNodeModifiers.IS_OUTPUT}}
    all_nodes = set()
    for n in fg.get_nodes():
        if n.user_defined:
            node_modifiers[n.name] = {graph.VisualizationNodeModifiers.IS_USER_INPUT}
        all_nodes.add(n)
    # hack of a test -- but it works... sort the lines and match them up.
    # why? because for some reason given the same graph, the output file isn't deterministic.
    # for the same reason, order of input nodes are non-deterministic
    expected_set = set(
        [
            '\t\tfunction [fillcolor="#b4d8e4" fontname=Helvetica margin=0.15 shape=rectangle style="rounded,filled"]\n',
            '\t\tgraph [fillcolor="#ffffff" fontname=helvetica label=Legend rank=same]\n',
            '\t\tinput [fontname=Helvetica margin=0.15 shape=rectangle style="filled,dashed"]\n',
            '\t\toutput [fillcolor="#FFC857" fontname=Helvetica margin=0.15 shape=rectangle style="rounded,filled"]\n',
            "\tA -> B\n",
            "\tA -> C\n",
            '\tA [label=<<b>A</b><br /><br /><i>int</i>> fillcolor="#b4d8e4" fontname=Helvetica margin=0.15 shape=rectangle style="rounded,filled"]\n',
            '\tB [label=<<b>B</b><br /><br /><i>int</i>> fillcolor="#FFC857" fontname=Helvetica margin=0.15 shape=rectangle style="rounded,filled"]\n',
            '\tC [label=<<b>C</b><br /><br /><i>int</i>> fillcolor="#b4d8e4" fontname=Helvetica margin=0.15 shape=rectangle style="rounded,filled"]\n',
            "\tb [label=<<b>b</b><br /><br /><i>1</i>> fontname=Helvetica shape=note style=filled]\n",
            "\tc [label=<<b>c</b><br /><br /><i>2</i>> fontname=Helvetica shape=note style=filled]\n",
            "\t_A_inputs -> A\n",
            # commenting out input node: '\t_A_inputs [label=<<table border="0"><tr><td>c</td><td>int</td></tr><tr><td>b</td><td>int</td></tr></table>> fontname=Helvetica margin=0.15 shape=rectangle style=dashed]\n',
            "\tgraph [compound=true concentrate=true rankdir=LR ranksep=0.4 style=filled]\n",
            '\tnode [fillcolor="#ffffff"]\n',
            "\tsubgraph cluster__legend {\n",
            "\t}\n",
            "// Dependency Graph\n",
            "digraph {\n",
            "}\n",
        ]
    )

    fg.display(
        all_nodes,
        output_file_path=str(dot_file_path),
        node_modifiers=node_modifiers,
        config=config,
        keep_dot=True,
    )
    dot_file = dot_file_path.open("r").readlines()
    dot_set = set(dot_file)

    assert dot_set.issuperset(expected_set) and len(dot_set.difference(expected_set)) == 1


@pytest.mark.parametrize(
    "filename, keep_dot", [("dag", False), ("dag.png", False), ("dag", True), ("dag.png", True)]
)
def test_function_graph_display_output_filename(
    tmp_path: pathlib.Path, filename: str, keep_dot: bool
):
    """Handle file generation with `graph.Digraph` `.render()` and `.pipe()`"""
    output_file_path = f"{tmp_path}/{filename}"
    config = {"b": 1, "c": 2}
    fg = graph.FunctionGraph.from_modules(tests.resources.dummy_functions, config=config)

    fg.display(
        set(fg.get_nodes()),
        output_file_path=output_file_path,
        config=config,
        keep_dot=keep_dot,
    )
    assert pathlib.Path(tmp_path, "dag.png").exists()
    assert pathlib.Path(tmp_path, "dag").exists() == keep_dot


def test_function_graph_display_no_dot_output(tmp_path: pathlib.Path):
    dot_file_path = tmp_path / "dag"
    config = {"b": 1, "c": 2}
    fg = graph.FunctionGraph.from_modules(tests.resources.dummy_functions, config=config)

    fg.display(set(fg.get_nodes()), output_file_path=None, config=config)

    assert not dot_file_path.exists()


def test_function_graph_display_custom_style_node():
    def _styling_function(*, node, node_class):
        return dict(fill_color="aquamarine"), None, "legend_key"

    config = {"b": 1, "c": 2}
    fg = graph.FunctionGraph.from_modules(tests.resources.dummy_functions, config=config)

    digraph = fg.display(
        set(fg.get_nodes()),
        custom_style_function=_styling_function,
        config=config,
    )

    key_found = False
    for line in digraph.body:  # list of lines of a dot file
        if ("A" in line) and ("aquamarine" in line):
            key_found = True
            break

    assert key_found


def test_function_graph_display_custom_style_legend():
    def _styling_function(*, node, node_class):
        return dict(fill_color="aquamarine"), None, "legend_key"

    config = {"b": 1, "c": 2}
    fg = graph.FunctionGraph.from_modules(tests.resources.dummy_functions, config=config)

    digraph = fg.display(
        set(fg.get_nodes()),
        custom_style_function=_styling_function,
        config=config,
    )

    key_found = False
    for line in digraph.body:  # list of lines of a dot file
        if ("legend_key" in line) and ("aquamarine" in line):
            key_found = True
            break

    assert key_found


def test_function_graph_display_custom_style_tag():
    def _styling_function(*, node, node_class):
        if node.tags.get("some_key"):
            style = dict(fill_color="aquamarine"), None, "legend_key"
        else:
            style = dict(), None, None
        return style

    nodes = create_testing_nodes()
    nodes["A"].tags["some_key"] = "some_value"
    config = {"b": 1, "c": 2}
    fg = graph.FunctionGraph(nodes, config=config)

    digraph = fg.display(
        set(fg.get_nodes()),
        custom_style_function=_styling_function,
        config=config,
    )

    # check that style is only applied to tagged nodes
    tag_found = False
    not_tag_found = False
    for line in digraph.body:  # list of lines of a dot file
        if ("A" in line) and ("aquamarine" in line):
            tag_found = True
        elif ("B" in line) and ("aquamarine" not in line):
            not_tag_found = True

    assert tag_found and not_tag_found


@pytest.mark.parametrize("show_legend", [(True), (False)])
def test_function_graph_display_legend(show_legend: bool):
    config = {"b": 1, "c": 2}
    fg = graph.FunctionGraph.from_modules(tests.resources.dummy_functions, config=config)

    dot = fg.display(
        set(fg.get_nodes()),
        show_legend=show_legend,
        config=config,
    )

    found_legend = "cluster__legend" in dot.source
    assert found_legend is show_legend


@pytest.mark.parametrize("orient", [("LR"), ("TB"), ("RL"), ("BT")])
def test_function_graph_display_orient(orient: str):
    config = {"b": 1, "c": 2}
    fg = graph.FunctionGraph.from_modules(tests.resources.dummy_functions, config=config)

    dot = fg.display(
        set(fg.get_nodes()),
        orient=orient,
        config=config,
    )

    # this could break if a rankdir is given to the legend subgraph
    assert f"rankdir={orient}" in dot.source


@pytest.mark.parametrize("hide_inputs", [(True,), (False,)])
def test_function_graph_display_inputs(hide_inputs: bool):
    config = {"b": 1, "c": 2}
    fg = graph.FunctionGraph.from_modules(tests.resources.dummy_functions, config=config)

    dot = fg.display(
        set(fg.get_nodes()),
        hide_inputs=hide_inputs,
        config=config,
    )

    found_input = any(line.startswith("\t_") for line in dot.body)
    assert found_input is not hide_inputs


def test_function_graph_display_without_saving():
    """Tests that display works when None is passed in for path"""
    config = {"b": 1, "c": 2}
    fg = graph.FunctionGraph.from_modules(tests.resources.dummy_functions, config=config)
    all_nodes = set()
    node_modifiers = {"B": {graph.VisualizationNodeModifiers.IS_OUTPUT}}
    for n in fg.get_nodes():
        if n.user_defined:
            node_modifiers[n.name] = {graph.VisualizationNodeModifiers.IS_USER_INPUT}
        all_nodes.add(n)
    digraph = fg.display(
        all_nodes, output_file_path=None, node_modifiers=node_modifiers, config=config
    )
    assert digraph is not None
    import graphviz

    assert isinstance(digraph, graphviz.Digraph)


@pytest.mark.parametrize("display_fields", [(True,), (False,)])
def test_function_graph_display_fields(display_fields: bool):
    @schema.output(("foo", "int"), ("bar", "float"), ("baz", "str"))
    def df_with_schema() -> pd.DataFrame:
        pass

    mod = ad_hoc_utils.create_temporary_module(df_with_schema)
    config = {}
    fg = graph.FunctionGraph.from_modules(mod, config=config)

    dot = fg.display(
        set(fg.get_nodes()),
        display_fields=display_fields,
        config=config,
    )
    if display_fields:
        assert any("foo" in line for line in dot.body)
        assert any("bar" in line for line in dot.body)
        assert any("baz" in line for line in dot.body)
        assert any("cluster" in line for line in dot.body)
    else:
        assert not any("foo" in line for line in dot.body)
        assert not any("bar" in line for line in dot.body)
        assert not any("baz" in line for line in dot.body)
        assert not any("cluster" in line for line in dot.body)


def test_function_graph_display_fields_shared_schema():
    # This ensures an edge case where they end up getting dropped if there are duplicates
    SCHEMA = (("foo", "int"), ("bar", "float"), ("baz", "str"))

    @schema.output(*SCHEMA)
    def df_1_with_schema() -> pd.DataFrame:
        pass

    @schema.output(*SCHEMA)
    def df_2_with_schema() -> pd.DataFrame:
        pass

    mod = ad_hoc_utils.create_temporary_module(df_1_with_schema, df_2_with_schema)
    config = {}
    fg = graph.FunctionGraph.from_modules(mod, config=config)

    dot = fg.display(
        set(fg.get_nodes()),
        display_fields=True,
        config=config,
    )

    def _get_occurances(var: str, lines: List[str]):
        return [item for item in lines if var in item]

    # We just need to make sure these show up twice
    assert len(_get_occurances("foo=", dot.body)) == 2
    assert len(_get_occurances("bar=", dot.body)) == 2
    assert len(_get_occurances("baz=", dot.body)) == 2


def test_function_graph_display_config_node():
    """Check if config is displayed by low-level hamilton.graph.FunctionGraph.display"""
    config = {"X": 1}
    fg = graph.FunctionGraph.from_modules(tests.resources.dummy_functions, config=config)

    dot = fg.display(set(fg.get_nodes()), config=config)

    # lines start tab then node name; check if "b" is a node in the graphviz object
    assert any(line.startswith("\tX") for line in dot.body)


# TODO use high-level visualization dot as fixtures for reuse across tests
def test_display_config_node():
    """Check if config is displayed by high-level hamilton.driver.display..."""
    from hamilton import driver
    from hamilton.io.materialization import to

    config = {"X": 1}
    dr = driver.Builder().with_modules(tests.resources.dummy_functions).with_config(config).build()

    all_dot = dr.display_all_functions()
    down_dot = dr.display_downstream_of("A")
    up_dot = dr.display_upstream_of("C")
    between_dot = dr.visualize_path_between("A", "C")
    exec_dot = dr.visualize_execution(["C"], inputs={"b": 1, "c": 2})
    materialize_dot = dr.visualize_materialization(
        to.json(id="saver", dependencies=["C"], combine=base.DictResult(), path="saver.json"),
        inputs={"b": 1, "c": 2},
    )

    for dot in [all_dot, down_dot, up_dot, between_dot, exec_dot, materialize_dot]:
        assert any(line.startswith("\tX") for line in dot.body)


def test_create_graphviz_graph():
    """Tests that we create a graphviz graph"""
    config = {}
    fg = graph.FunctionGraph.from_modules(tests.resources.dummy_functions, config=config)
    nodes, user_nodes = fg.get_upstream_nodes(["A", "B", "C"])
    nodez = nodes.union(user_nodes)
    node_modifiers = {
        "b": {graph.VisualizationNodeModifiers.IS_USER_INPUT},
        "c": {graph.VisualizationNodeModifiers.IS_USER_INPUT},
        "B": {graph.VisualizationNodeModifiers.IS_OUTPUT},
    }
    # hack of a test -- but it works... sort the lines and match them up.
    # why? because for some reason given the same graph, the output file isn't deterministic.
    # for the same reason, order of input nodes are non-deterministic
    expected_set = set(
        [
            "// Dependency Graph",
            "",
            "digraph {",
            '\tnode [fillcolor="#ffffff"]',
            "\tgraph [compound=true concentrate=true rankdir=LR ranksep=0.4 ratio=1 style=filled]",
            '\tB [label=<<b>B</b><br /><br /><i>int</i>> fillcolor="#FFC857" fontname=Helvetica margin=0.15 shape=rectangle style="rounded,filled"]',
            '\tC [label=<<b>C</b><br /><br /><i>int</i>> fillcolor="#b4d8e4" fontname=Helvetica margin=0.15 shape=rectangle style="rounded,filled"]',
            '\tA [label=<<b>A</b><br /><br /><i>int</i>> fillcolor="#b4d8e4" fontname=Helvetica margin=0.15 shape=rectangle style="rounded,filled"]',
            "\tA -> B",
            "\tA -> C",
            '\t_A_inputs [label=<<table border="0"><tr><td>b</td><td>int</td></tr><tr><td>b</td><td>int</td></tr></table>> fontname=Helvetica margin=0.15 shape=rectangle style="filled,dashed"]',
            "\t_A_inputs -> A",
            "\tsubgraph cluster__legend {",
            '\t\tgraph [fillcolor="#ffffff" fontname=helvetica label=Legend rank=same]',
            '\t\tinput [fontname=Helvetica margin=0.15 shape=rectangle style="filled,dashed"]',
            '\t\tfunction [fillcolor="#b4d8e4" fontname=Helvetica margin=0.15 shape=rectangle style="rounded,filled"]',
            '\t\toutput [fillcolor="#FFC857" fontname=Helvetica margin=0.15 shape=rectangle style="rounded,filled"]',
            "\t}",
            "}",
            "",
        ]
    )

    digraph = graph.create_graphviz_graph(
        nodez,
        "Dependency Graph\n",
        graphviz_kwargs=dict(graph_attr={"ratio": "1"}),
        node_modifiers=node_modifiers,
        strictly_display_only_nodes_passed_in=False,
        config=config,
    )
    # the HTML table isn't deterministic. Replace the value in it with a single one.
    dot_set = set(str(digraph).replace("<td>c</td>", "<td>b</td>").split("\n"))
    assert dot_set == expected_set


def test_create_networkx_graph():
    """Tests that we create a networkx graph"""
    fg = graph.FunctionGraph.from_modules(tests.resources.dummy_functions, config={})
    nodes, user_nodes = fg.get_upstream_nodes(["A", "B", "C"])
    digraph = graph.create_networkx_graph(nodes, user_nodes, "test-graph")
    expected_nodes = sorted(["c", "B", "C", "b", "A"])
    expected_edges = sorted([("c", "A"), ("b", "A"), ("A", "B"), ("A", "C")])
    assert sorted(list(digraph.nodes)) == expected_nodes
    assert sorted(list(digraph.edges)) == expected_edges


def test_end_to_end_with_layered_decorators_resolves_true():
    fg = graph.FunctionGraph.from_modules(
        tests.resources.layered_decorators, config={"foo": "bar", "d": 10, "b": 20}
    )
    out = fg.execute([n for n in fg.get_nodes()])
    assert len(out) > 0  # test config.when resolves correctly
    assert out["e"] == (20 + 10)
    assert out["f"] == (20 + 20)


def test_end_to_end_with_layered_decorators_resolves_false():
    config = {"foo": "not_bar", "d": 10, "b": 20}
    fg = graph.FunctionGraph.from_modules(tests.resources.layered_decorators, config=config)
    out = fg.execute(
        [n for n in fg.get_nodes()],
    )
    assert {item: value for item, value in out.items() if item not in config} == {}


def test_combine_inputs_no_collision():
    """Tests the combine_and_validate_inputs functionality when there are no collisions"""
    combined = graph_functions.combine_config_and_inputs({"a": 1}, {"b": 2})
    assert combined == {"a": 1, "b": 2}


def test_combine_inputs_collision():
    """Tests the combine_and_validate_inputs functionality
    when there are collisions of keys but not values"""
    with pytest.raises(ValueError):
        graph_functions.combine_config_and_inputs({"a": 1}, {"a": 2})


def test_combine_inputs_collision_2():
    """Tests the combine_and_validate_inputs functionality
    when there are collisions of keys and values"""
    with pytest.raises(ValueError):
        graph_functions.combine_config_and_inputs({"a": 1}, {"a": 1})


def test_extract_columns_executes_once():
    """Ensures that extract_columns only computes the function once.
    Note this is a bit heavy-handed of a test but its nice to have."""
    fg = graph.FunctionGraph.from_modules(
        tests.resources.extract_columns_execution_count, config={}
    )
    unique_id = str(uuid.uuid4())
    fg.execute([n for n in fg.get_nodes()], inputs={"unique_id": unique_id})
    assert (
        len(tests.resources.extract_columns_execution_count.outputs[unique_id]) == 1
    )  # It should only be called once


def test_end_to_end_with_generics():
    fg = graph.FunctionGraph.from_modules(
        tests.resources.functions_with_generics, config={"b": {}, "c": 1}
    )
    results = fg.execute(fg.get_nodes())
    assert results == {
        "A": ({}, 1),
        "B": [({}, 1), ({}, 1)],
        "C": {"foo": [({}, 1), ({}, 1)]},
        "D": 1.0,
        "b": {},
        "c": 1,
    }


@pytest.mark.parametrize(
    "config,inputs,overrides",
    [
        # testing with no provided inputs
        ({}, {}, {}),
        # testing with just configs
        ({"a": 11}, {}, {}),
        ({"b": 13, "a": 17}, {}, {}),
        ({"b": 19, "a": 23, "d": 29, "f": 31}, {}, {}),
        # Testing with just inputs
        ({}, {"a": 37}, {}),
        ({}, {"b": 41, "a": 43}, {}),
        ({}, {"b": 41, "a": 43, "d": 47, "f": 53}, {}),
        # Testing with just overrides
        # TBD whether these should be legitimate -- can we override required inputs?
        # Test works now but not part of the contract
        # ({}, {}, {'a': 59}),
        # ({}, {}, {'a': 61, 'b': 67}),
        # ({}, {}, {'a': 71, 'b': 73, 'd': 79, 'f': 83}),
        # testing with a mix
        ({"a": 89}, {"b": 97}, {}),
        ({"a": 101}, {"b": 103, "d": 107}, {}),
    ],
)
def test_optional_execute(config, inputs, overrides):
    """Tests execution of optionals with different assortment of overrides, configs, inputs, etc...
    Be careful adding tests with conflicting values between them.
    """
    fg = graph.FunctionGraph.from_modules(tests.resources.optional_dependencies, config=config)
    # we put a user input node first to ensure that order does not matter with computation order.
    results = fg.execute([fg.nodes["b"], fg.nodes["g"]], inputs=inputs, overrides=overrides)
    do_all_args = {key + "_val": val for key, val in {**config, **inputs, **overrides}.items()}
    expected_results = tests.resources.optional_dependencies._do_all(**do_all_args)
    assert results["g"] == expected_results["g"]


def test_optional_input_behavior():
    """Tests that if we request optional user inputs that are not provided, we do not break. And if they are
    we do the right thing and return them.
    """
    fg = graph.FunctionGraph.from_modules(tests.resources.optional_dependencies, config={})
    # nothing passed, so nothing returned
    result = fg.execute([fg.nodes["b"], fg.nodes["a"]], inputs={}, overrides={})
    assert result == {}
    # something passed, something returned via config
    fg2 = graph.FunctionGraph.from_modules(tests.resources.optional_dependencies, config={"a": 10})
    result = fg2.execute([fg.nodes["b"], fg.nodes["a"]], inputs={}, overrides={})
    assert result == {"a": 10}
    # something passed, something returned via inputs
    result = fg.execute([fg.nodes["b"], fg.nodes["a"]], inputs={"a": 10}, overrides={})
    assert result == {"a": 10}
    # something passed, something returned via overrides
    result = fg.execute([fg.nodes["b"], fg.nodes["a"]], inputs={}, overrides={"a": 10})
    assert result == {"a": 10}


@pytest.mark.parametrize("node_order", list(permutations("fhi")))
def test_user_input_breaks_if_required_missing(node_order):
    """Tests that we break because `h` is required but is not passed in."""
    fg = graph.FunctionGraph.from_modules(tests.resources.optional_dependencies, config={})
    permutation = [fg.nodes[n] for n in node_order]
    with pytest.raises(NotImplementedError):
        fg.execute(permutation, inputs={}, overrides={})


@pytest.mark.parametrize("node_order", list(permutations("fhi")))
def test_user_input_does_not_break_if_required_provided(node_order):
    """Tests that things work no matter the order because `h` is required and is passed in, while `f` is optional."""
    fg = graph.FunctionGraph.from_modules(tests.resources.optional_dependencies, config={"h": 10})
    permutation = [fg.nodes[n] for n in node_order]
    result = fg.execute(permutation, inputs={}, overrides={})
    assert result == {"h": 10, "i": 17}


def test_optional_donot_drop_none():
    """We do not want to drop `None` results from functions. We want to pass them through to the function.

    This is here to enshrine the current behavior.
    """
    fg = graph.FunctionGraph.from_modules(tests.resources.optional_dependencies, config={"h": None})
    # enshrine behavior that None is not removed from being passed to the function.
    results = fg.execute([fg.nodes["h"], fg.nodes["i"]], inputs={}, overrides={})
    assert results == {"h": None, "i": 17}
    fg = graph.FunctionGraph.from_modules(tests.resources.optional_dependencies, config={})
    results = fg.execute(
        [fg.nodes["j"], fg.nodes["none_result"], fg.nodes["f"]], inputs={}, overrides={}
    )
    assert results == {"j": None, "none_result": None}  # f omitted cause it's optional.


def test_optional_get_required_compile_time():
    """Tests that getting required with optionals at compile time returns everything
    TODO -- change this to be testing a different function (compile time) than runtime.
    """
    fg = graph.FunctionGraph.from_modules(tests.resources.optional_dependencies, config={})
    all_upstream, user_required = fg.get_upstream_nodes(["g"])
    assert len(all_upstream) == 7  # 6 total nodes upstream
    assert len(user_required) == 4  # 4 nodes required input


def test_optional_get_required_runtime():
    fg = graph.FunctionGraph.from_modules(tests.resources.optional_dependencies, config={})
    all_upstream, user_required = fg.get_upstream_nodes(["g"], runtime_inputs={})  # Nothng required
    assert len(all_upstream) == 3  # 6 total nodes upstream
    assert len(user_required) == 0  # 4 nodes required input


def test_optional_get_required_runtime_with_provided():
    fg = graph.FunctionGraph.from_modules(tests.resources.optional_dependencies, config={})
    all_upstream, user_required = fg.get_upstream_nodes(
        ["g"], runtime_inputs={"b": 109}
    )  # Nothng required
    assert len(all_upstream) == 4  # 6 total nodes upstream
    assert len(user_required) == 1  # 4 nodes required input


def test_in_driver_function_definitions():
    """Tests that we can instantiate a DAG with a function defined in the driver, e.g. notebook context"""

    def my_function(A: int, b: int, c: int) -> int:
        """Function for input below"""
        return A + b + c

    f_module = ad_hoc_utils.create_temporary_module(my_function)
    fg = graph.FunctionGraph.from_modules(
        tests.resources.dummy_functions, f_module, config={"b": 3, "c": 1}
    )
    results = fg.execute([n for n in fg.get_nodes() if n.name in ["my_function", "A"]])
    assert results == {"A": 4, "b": 3, "c": 1, "my_function": 8}


def test_update_dependencies():
    nodes = create_testing_nodes()
    new_nodes = graph.update_dependencies(
        nodes, lifecycle_base.LifecycleAdapterSet(base.DefaultAdapter())
    )
    for node_name, node_ in new_nodes.items():
        assert node_.dependencies == nodes[node_name].dependencies
        assert node_.depended_on_by == nodes[node_name].depended_on_by



from typing import Any, Type

from hamilton.data_quality import base, default_validators
from hamilton.htypes import custom_subclass_check

try:
    import pydantic  # noqa: F401
except ModuleNotFoundError as e:
    raise NotImplementedError(
        "Cannot import `pydantic` from `pydantic_validators`. Run pip install 'sf-hamilton[pydantic]' if needed."
    ) from e

try:
    from pydantic import BaseModel, TypeAdapter, ValidationError
except ImportError as e:
    raise NotImplementedError(
        "`pydantic>=2.0` required to use `pydantic_validators`. Run pip install 'sf-hamilton[pydantic]' if needed."
    ) from e


COLUMN_FRIENDLY_DF_TYPE = False


class PydanticModelValidator(base.BaseDefaultValidator):
    """Pydantic model compatibility validator (requires ``pydantic>=2.0``)

    Note that this validator uses pydantic's strict mode, which does not allow for
    coercion of data. This means that if an object does not exactly match the reference
    type, it will fail validation, regardless of whether it could be coerced into the
    correct type.

    :param model: Pydantic model to validate against
    :param importance: Importance of the validator, possible values "warn" and "fail"
    :param arbitrary_types_allowed: Whether arbitrary types are allowed in the model
    """

    def __init__(self, model: Type[BaseModel], importance: str):
        super(PydanticModelValidator, self).__init__(importance)
        self.model = model
        self._model_adapter = TypeAdapter(model)

    @classmethod
    def applies_to(cls, datatype: Type[Type]) -> bool:
        # In addition to checking for a subclass of BaseModel, we also check for dict
        # as this is the standard 'de-serialized' format of pydantic models in python
        return custom_subclass_check(datatype, BaseModel) or custom_subclass_check(datatype, dict)

    def description(self) -> str:
        return "Validates that the returned object is compatible with the specified pydantic model"

    def validate(self, data: Any) -> base.ValidationResult:
        try:
            # Currently, validate can not alter the output data, so we must use
            # strict=True. The downside to this is that data that could be coerced
            # into the correct type will fail validation.
            self._model_adapter.validate_python(data, strict=True)
        except ValidationError as e:
            return base.ValidationResult(
                passes=False, message=str(e), diagnostics={"model_errors": e.errors()}
            )
        return base.ValidationResult(
            passes=True,
            message=f"Data passes pydantic check for model {str(self.model)}",
        )

    @classmethod
    def arg(cls) -> str:
        return "model"

    @classmethod
    def name(cls) -> str:
        return "pydantic_validator"


def register_validators():
    """Utility method to append pydantic validators as needed"""
    validators = [PydanticModelValidator]
    default_validators.AVAILABLE_DEFAULT_VALIDATORS.extend(validators)


register_validators()



import dataclasses
from pathlib import Path
from typing import Any, Collection, Dict, Literal, Optional, Tuple, Type, Union

try:
    import lightgbm
except ImportError as e:
    raise NotImplementedError("LightGBM is not installed.") from e


from hamilton import registry
from hamilton.io import utils
from hamilton.io.data_adapters import DataLoader, DataSaver

LIGHTGBM_MODEL_TYPES = [lightgbm.LGBMModel, lightgbm.Booster, lightgbm.CVBooster]
LIGHTGBM_MODEL_TYPES_ANNOTATION = Union[lightgbm.LGBMModel, lightgbm.Booster, lightgbm.CVBooster]


@dataclasses.dataclass
class LightGBMFileWriter(DataSaver):
    """Write LighGBM models and boosters to a file"""

    path: Union[str, Path]
    num_iteration: Optional[int] = None
    start_iteration: int = 0
    importance_type: Literal["split", "gain"] = "split"

    @classmethod
    def applicable_types(cls) -> Collection[Type]:
        return LIGHTGBM_MODEL_TYPES

    def save_data(self, data: LIGHTGBM_MODEL_TYPES_ANNOTATION) -> Dict[str, Any]:
        if isinstance(data, lightgbm.LGBMModel):
            data = data.booster_

        data.save_model(
            filename=self.path,
            num_iteration=self.num_iteration,
            start_iteration=self.start_iteration,
            importance_type=self.importance_type,
        )
        return utils.get_file_metadata(self.path)

    @classmethod
    def name(cls) -> str:
        return "file"


@dataclasses.dataclass
class LightGBMFileReader(DataLoader):
    """Load LighGBM models and boosters from a file"""

    path: Union[str, Path]

    @classmethod
    def applicable_types(cls) -> Collection[Type]:
        return LIGHTGBM_MODEL_TYPES

    def load_data(
        self, type_: Type
    ) -> Tuple[Union[lightgbm.Booster, lightgbm.CVBooster], Dict[str, Any]]:
        model = type_(model_file=self.path)
        metadata = utils.get_file_metadata(self.path)
        return model, metadata

    @classmethod
    def name(cls) -> str:
        return "file"


def register_data_loaders():
    for loader in [
        LightGBMFileReader,
        LightGBMFileWriter,
    ]:
        registry.register_adapter(loader)


register_data_loaders()

COLUMN_FRIENDLY_DF_TYPE = False



from typing import Any, Dict, Type, Union

import polars as pl

from hamilton import base


class PolarsDataFrameResult(base.ResultMixin):
    """A ResultBuilder that produces a polars dataframe.

    Use this when you want to create a polars dataframe from the outputs. Caveat: you need to ensure that the length
    of the outputs is the same, otherwise you will get an error; mixed outputs aren't that well handled.

    To use:

    .. code-block:: python

        from hamilton import base, driver
        from hamilton.plugins import polars_extensions

        polars_builder = polars_extensions.PolarsDataFrameResult()
        adapter = base.SimplePythonGraphAdapter(polars_builder)
        dr = driver.Driver(config, *modules, adapter=adapter)
        df = dr.execute([...], inputs=...)  # returns polars dataframe

    Note: this is just a first attempt at something for Polars. Think it should handle more? Come chat/open a PR!
    """

    def build_result(
        self, **outputs: Dict[str, Union[pl.Series, pl.DataFrame, Any]]
    ) -> pl.DataFrame:
        """This is the method that Hamilton will call to build the final result. It will pass in the results
        of the requested outputs that you passed in to the execute() method.

        Note: this function could do smarter things; looking for contributions here!

        :param outputs: The results of the requested outputs.
        :return: a polars DataFrame.
        """
        if len(outputs) == 1:
            (value,) = outputs.values()  # this works because it's length 1.
            if isinstance(value, pl.DataFrame):  # it's a dataframe
                return value
            if isinstance(value, pl.LazyFrame):  # it's a lazyframe
                return value.collect()
            elif not isinstance(value, pl.Series):  # it's a single scalar/object
                key, value = outputs.popitem()
                return pl.DataFrame({key: [value]})
            else:  # it's a series
                return pl.DataFrame(outputs)
        # TODO: check for length of outputs and determine what should
        # happen for mixed outputs that include scalars for example.
        return pl.DataFrame(outputs)

    def output_type(self) -> Type:
        return pl.DataFrame



import collections
import sys
import typing
from typing import Any, Dict, List, Union

import pandas as pd
import pytest

from hamilton import htypes
from hamilton.htypes import check_instance


class X:
    pass


class Y(X):
    pass


custom_type = typing.TypeVar("FOOBAR")


@pytest.mark.parametrize(
    "param_type,requested_type,expected",
    [
        (custom_type, custom_type, True),
        (custom_type, typing.TypeVar("FOO"), False),
        (typing.Any, typing.TypeVar("FOO"), True),
        (typing.Any, custom_type, True),
        (int, int, True),
        (int, float, False),
        (typing.List[int], typing.List, True),
        (typing.List, typing.List[float], True),
        (typing.List, list, True),
        (typing.Dict, dict, True),
        (dict, typing.Dict, True),
        (list, typing.List, True),
        (list, typing.List, True),
        (typing.List[int], typing.List[float], False),
        (typing.Dict, typing.List, False),
        (typing.Mapping, typing.Dict, True),
        (typing.Mapping, dict, True),
        (dict, typing.Mapping, False),
        (typing.Dict, typing.Mapping, False),
        (typing.Iterable, typing.List, True),
        (typing.Tuple[str, str], typing.Tuple[str, str], True),
        (typing.Tuple[str, str], typing.Tuple[str], False),
        (typing.Tuple[str, str], typing.Tuple, True),
        (typing.Tuple, typing.Tuple[str, str], True),
        (typing.Union[str, str], typing.Union[str, str], True),
        (X, X, True),
        (X, Y, True),
        (Y, X, False),
        (typing.Any, Y, True),
        (Y, typing.Any, False),
        (typing.Union[X, int], X, True),
        (typing.Union[str, X], str, True),
        (typing.Union[custom_type, X], Y, True),
        (typing.Union[float, str], int, False),
        (typing.Union[int, float], X, False),
        (collections.Counter, collections.Counter, True),
        (dict, collections.Counter, True),
        (typing.Dict, collections.Counter, True),
        # These are not subclasses of each other, see issue 42
        (typing.FrozenSet[int], typing.Set[int], False),
        (htypes.column[pd.Series, int], pd.Series, True),
        (htypes.column[pd.Series, int], int, False),
        (typing.Any, pd.DataFrame, True),
        (pd.DataFrame, typing.Any, False),
    ],
)
def test_custom_subclass_check(param_type, requested_type, expected):
    """Tests the custom_subclass_check"""
    actual = htypes.custom_subclass_check(requested_type, param_type)
    assert actual == expected


@pytest.mark.parametrize(
    "param_type,required_type,expected",
    [
        (typing.TypeVar("FOO"), typing.TypeVar("BAR"), False),
        (custom_type, custom_type, True),
        (int, int, True),
        (int, float, False),
        (typing.Dict, typing.Any, True),
        (X, X, True),
        (X, Y, True),
        (pd.Series, pd.Series, True),
        (list, pd.Series, False),
        (dict, pd.Series, False),
    ],
)
def test_types_match(param_type, required_type, expected):
    """Tests the types_match function"""
    actual = htypes.types_match(param_type, required_type)
    assert actual == expected


@pytest.mark.parametrize(
    "type_",
    [
        int,
        bool,
        float,
        pd.Series,
        pd.DataFrame,
        htypes.column[pd.Series, int],
        htypes.column[pd.Series, float],
        htypes.column[pd.Series, bool],
        htypes.column[pd.Series, str],
    ],
)
def test_validate_types_happy(type_):
    """Tests that validate_types works when the type is valid"""
    htypes.validate_type_annotation(type_)


@pytest.mark.parametrize(
    "type_",
    [
        htypes.column[pd.DataFrame, int],
        htypes.column[pd.DataFrame, float],
        htypes.column[pd.Series, typing.Dict[str, typing.Any]],
    ],
)
def test_validate_types_sad(type_):
    """Tests that validate_types works when the type is valid"""
    with pytest.raises(htypes.InvalidTypeException):
        htypes.validate_type_annotation(type_)


@pytest.mark.parametrize(
    "candidate,type_,expected",
    [
        (int, int, True),
        (int, float, False),
        # Not safe so we return false
        (typing.List[int], typing.List, False),
        (typing.FrozenSet[int], typing.Set[int], False),
        (typing.Dict, dict, False),
    ],
)
def test__safe_subclass(candidate, type_, expected):
    assert htypes._safe_subclass(candidate, type_) == expected


@pytest.mark.parametrize(
    "type_",
    [
        (custom_type),
        (typing.TypeVar("FOO")),
        (typing.Any),
        (int),
        (float),
        (typing.List[int]),
        (typing.List),
        (list),
        (typing.Iterable),
        (typing.Dict),
        (dict),
        (typing.Mapping),
        (collections.Counter),
        (typing.Tuple[str, str]),
        (typing.Tuple[str]),
        (typing.Tuple),
        (typing.Union[str, str]),
        (X),
        (Y),
        (typing.Any),
        (typing.Union[X, int]),
        (typing.Union[str, X]),
        (typing.Union[custom_type, X]),
        (typing.Union[float, str]),
        (typing.Union[int, float]),
        (typing.FrozenSet[int]),
        (typing.Set[int]),
        (pd.Series),
        (htypes.column[pd.Series, int]),
        (pd.DataFrame),
    ],
)
def test_get_type_as_string(type_):
    """Tests the custom_subclass_check"""
    try:
        type_string = htypes.get_type_as_string(type_)  # noqa: F841
    except Exception as e:
        pytest.fail(f"test get_type_as_string raised: {e}")


@pytest.mark.parametrize(
    "node_type,input_value",
    [
        (pd.DataFrame, pd.Series([1, 2, 3])),
        (typing.List, {}),
        (typing.Dict, []),
        (dict, []),
        (list, {}),
        (int, 1.0),
        (float, 1),
        (str, 0),
        (typing.Union[int, pd.Series], pd.DataFrame({"a": [1, 2, 3]})),
        (typing.Union[int, pd.Series], 1.0),
        (typing.Sequence, {"a", "b"}),
    ],
    ids=[
        "test-subclass",
        "test-generic-list",
        "test-generic-dict",
        "test-type-match-dict",
        "test-type-match-list",
        "test-type-match-int",
        "test-type-match-float",
        "test-type-match-str",
        "test-union-mismatch-dataframe",
        "test-union-mismatch-float",
        "test-sequence-set-mismatch",
    ],
)
def test_check_input_type_mismatch(node_type, input_value):
    """Tests check_input_type of SimplePythonDataFrameGraphAdapter"""
    actual = htypes.check_input_type(node_type, input_value)
    assert actual is False


T = typing.TypeVar("T")


@pytest.mark.parametrize(
    "node_type,input_value",
    [
        (typing.Any, None),
        (pd.Series, pd.Series([1, 2, 3])),
        (T, None),
        (typing.List, []),
        (typing.Dict, {}),
        (dict, {}),
        (list, []),
        (int, 1),
        (float, 1.0),
        (str, "abc"),
        (typing.Union[int, pd.Series], pd.Series([1, 2, 3])),
        (typing.Union[int, pd.Series], 1),
        (typing.Literal["csv", "prq"], "csv"),
        (typing.Sequence[str], ("a", "b")),
        (typing.Sequence, ("a", "b")),
        (typing.Iterable, ("a", "b")),
        (typing.Iterable[str], ["a", "b"]),
        (typing.Iterable[str], {"a", "b"}),
        (typing.Iterable[int], range(0, 10)),
        (typing.Sequence[int], range(0, 10)),
    ],
    ids=[
        "test-any",
        "test-subclass",
        "test-typevar",
        "test-generic-list",
        "test-generic-dict",
        "test-type-match-dict",
        "test-type-match-list",
        "test-type-match-int",
        "test-type-match-float",
        "test-type-match-str",
        "test-union-match-series",
        "test-union-match-int",
        "test-literal-match-str",
        "test-sequence-str-match-tuple-str",
        "test-sequence-match-tuple-str",
        "test-iterable-str-match-tuple-str",
        "test-iterable-str-match-list-str",
        "test-iterable-str-match-set-str",
        "test-iterable-int-match-range",
        "test-sequence-int-match-range",
    ],
)
def test_check_input_type_match(node_type, input_value):
    """Tests check_input_type of SimplePythonDataFrameGraphAdapter"""
    actual = htypes.check_input_type(node_type, input_value)
    assert actual is True


# We cannot parameterize this as the parameterization cannot be
# included if the
@pytest.mark.skipif(
    sys.version_info < (3, 9, 0),
    reason="Type hinting generics in standard collections " "is only supported in 3.9+",
)
def test_check_input_types_subscripted_generics_dict_str_Any():
    """Tests check_input_type of SimplePythonDataFrameGraphAdapter"""
    actual = htypes.check_input_type(dict[str, typing.Any], {})
    assert actual is True


# We cannot parameterize this as the parameterization cannot be
# included if the
@pytest.mark.skipif(
    sys.version_info < (3, 9, 0),
    reason="Type hinting generics in standard collections " "is only supported in 3.9+",
)
def test_check_input_types_subscripted_generics_list_Any():
    """Tests check_input_type of SimplePythonDataFrameGraphAdapter"""
    actual = htypes.check_input_type(list[typing.Any], [])
    assert actual is True


def test_check_instance_with_non_generic_type():
    assert check_instance(5, int)
    assert not check_instance("5", int)


def test_check_instance_with_generic_list_type():
    assert check_instance([1, 2, 3], List[int])
    assert not check_instance([1, 2, "3"], List[int])
    if sys.version_info >= (3, 9):
        # skip 3.8 -- not worth fixing
        assert check_instance([1, 2, 3], List)
        assert check_instance([1, 2, "3"], List)


def test_check_instance_with_list_type():
    assert check_instance([1, 2, 3], list)
    assert check_instance([1, 2, "3"], list)


def test_check_instance_with_generic_dict_type():
    assert check_instance({"key1": 1, "key2": 2}, Dict[str, int])
    assert not check_instance({"key1": 1, "key2": "2"}, Dict[str, int])
    if sys.version_info >= (3, 9):
        # skip 3.8 -- not worth fixing
        assert check_instance({"key1": 1, "key2": 2}, Dict)
        assert check_instance({"key1": 1, "key2": "2"}, Dict)


def test_check_instance_with_dict_type():
    assert check_instance({"key1": 1, "key2": 2}, dict)
    assert check_instance({"key1": 1, "key2": "2"}, dict)


def test_check_instance_with_nested_generic_type():
    assert check_instance([{"key1": 1, "key2": 2}, {"key3": 3, "key4": 4}], List[Dict[str, int]])
    assert not check_instance(
        [{"key1": 1, "key2": 2}, {"key3": 3, "key4": "4"}], List[Dict[str, int]]
    )


def test_check_instance_with_none_type():
    assert check_instance(None, type(None))
    assert not check_instance(5, type(None))


def test_check_instance_with_any_type():
    assert check_instance(5, Any)
    assert check_instance("5", Any)
    assert check_instance([1, 2, 3], Any)
    assert check_instance({"key1": 1, "key2": 2}, Any)


def test_check_instance_with_union_type():
    assert check_instance(5, Union[int, str])
    assert check_instance("5", Union[int, str])
    assert not check_instance([1, 2, 3], Union[int, str])
    assert not check_instance({"key1": 1, "key2": 2}, Union[int, str])


@pytest.mark.skipif(sys.version_info < (3, 9), reason="requires python 3.9 or higher")
def test_check_instance_with_union_type_and_literal():
    from typing import Literal

    assert check_instance("a", Union[Literal["a"], Literal["b"]])
    assert check_instance("b", Union[Literal["a"], Literal["b"]])
    assert not check_instance("c", Union[Literal["a"], Literal["b"]])


@pytest.mark.skipif(sys.version_info < (3, 9), reason="requires python 3.9 or higher")
def test_non_generic_dict_and_list():
    assert check_instance([1, 2, 3], list[int])
    assert not check_instance([1, 2, "3"], list[int])
    assert check_instance({"key1": 1, "key2": 2}, dict[str, int])
    assert not check_instance({"key1": 1, "key2": "2"}, dict[str, int])


def test_with_random_object():
    class RandomObject:
        pass

    assert check_instance(RandomObject(), RandomObject)
    assert not check_instance(RandomObject(), int)
    assert not check_instance(RandomObject(), list)
    assert not check_instance(RandomObject(), dict)
    assert check_instance(RandomObject(), Any)
    assert check_instance(RandomObject(), Union[RandomObject, int])
    assert check_instance(RandomObject(), Union[RandomObject, int, str])
    assert not check_instance(RandomObject(), Union[int, str])
    assert not check_instance(RandomObject(), Union[int, str, list])
    assert not check_instance(RandomObject(), Union[int, str, dict])
    assert not check_instance(RandomObject(), Union[int, str, list, dict])
    assert not check_instance(RandomObject(), Union[int, str, list, dict, None])
    assert check_instance(RandomObject(), Union[int, str, list, dict, RandomObject])
    assert not check_instance(RandomObject(), Union[int, str, list, dict, None, float])
    assert check_instance(RandomObject(), Union[int, str, list, dict, None, RandomObject])
    assert check_instance(RandomObject(), Union[int, str, list, dict, None, RandomObject, float])
    assert not check_instance(RandomObject(), Union[int, str, list, dict, None, float, bool])
    assert check_instance(
        RandomObject(), Union[int, str, list, dict, None, RandomObject, float, bool]
    )
    assert not check_instance(RandomObject(), Union[int, str, list, dict, None, float, bool, bytes])
    assert check_instance(
        RandomObject(), Union[int, str, list, dict, None, RandomObject, float, bool, bytes]
    )
    assert not check_instance(
        RandomObject(), Union[int, str, list, dict, None, float, bool, bytes, complex]
    )
    assert check_instance(
        RandomObject(), Union[int, str, list, dict, None, RandomObject, float, bool, bytes, complex]
    )
    assert not check_instance(
        RandomObject(), Union[int, str, list, dict, None, float, bool, bytes, complex]
    )
    assert check_instance(
        RandomObject(), Union[int, str, list, dict, None, RandomObject, float, bool, bytes, complex]
    )
    assert not check_instance(
        RandomObject(), Union[int, str, list, dict, None, float, bool, bytes, complex, type(None)]
    )



try:
    import yaml
except ImportError as e:
    raise NotImplementedError("yaml is not installed and is needed for yaml hamilton plugin") from e

import dataclasses
import pathlib
from typing import Any, Collection, Dict, Tuple, Type, Union

from hamilton import registry
from hamilton.io.data_adapters import DataLoader, DataSaver
from hamilton.io.utils import get_file_metadata

PrimitiveTypes = str, int, float, bool, dict, list
AcceptedTypes = Union[PrimitiveTypes]


@dataclasses.dataclass
class YAMLDataLoader(DataLoader):
    path: Union[str, pathlib.Path]

    @classmethod
    def applicable_types(cls) -> Collection[Type]:
        return [*PrimitiveTypes]

    @classmethod
    def name(cls) -> str:
        return "yaml"

    def load_data(self, type_: Type) -> Tuple[AcceptedTypes, Dict[str, Any]]:
        path = self.path
        if isinstance(self.path, str):
            path = pathlib.Path(self.path)

        with path.open(mode="r") as f:
            return yaml.safe_load(f), get_file_metadata(path)


@dataclasses.dataclass
class YAMLDataSaver(DataSaver):
    path: Union[str, pathlib.Path]

    @classmethod
    def applicable_types(cls) -> Collection[Type]:
        return [*PrimitiveTypes]

    @classmethod
    def name(cls) -> str:
        return "yaml"

    def save_data(self, data: AcceptedTypes) -> Dict[str, Any]:
        path = self.path
        if isinstance(path, str):
            path = pathlib.Path(path)
        with path.open("w") as f:
            yaml.dump(data, f)
        return get_file_metadata(self.path)


COLUMN_FRIENDLY_DF_TYPE = False


def register_data_loaders():
    for materializer in [
        YAMLDataLoader,
        YAMLDataSaver,
    ]:
        registry.register_adapter(materializer)


register_data_loaders()



import shutil
from pathlib import Path
from typing import Any, Optional

try:
    from typing import override
except ImportError:
    override = lambda x: x  # noqa E731

from hamilton.io.data_adapters import DataLoader, DataSaver

from .base import ResultStore, StoredResult


class FileResultStore(ResultStore):
    def __init__(self, path: str, create_dir: bool = True) -> None:
        self.path = Path(path)
        self.create_dir = create_dir

        if self.create_dir:
            self.path.mkdir(exist_ok=True, parents=True)

    def __getstate__(self) -> dict:
        """Serialize the `__init__` kwargs to pass in Parallelizable branches
        when using multiprocessing.
        """
        return {"path": str(self.path)}

    @staticmethod
    def _write_result(file_path: Path, stored_result: StoredResult) -> None:
        file_path.write_bytes(stored_result.save())

    @staticmethod
    def _load_result_from_path(path: Path) -> Optional[StoredResult]:
        try:
            data = path.read_bytes()
            return StoredResult.load(data)
        except FileNotFoundError:
            return None

    def _path_from_data_version(self, data_version: str) -> Path:
        return self.path.joinpath(data_version)

    def _materialized_path(self, data_version: str, saver_cls: DataSaver) -> Path:
        # TODO allow a more flexible mechanism to specify file path extension
        return self._path_from_data_version(data_version).with_suffix(f".{saver_cls.name()}")

    @override
    def exists(self, data_version: str) -> bool:
        result_path = self._path_from_data_version(data_version)
        return result_path.exists()

    @override
    def set(
        self,
        data_version: str,
        result: Any,
        saver_cls: Optional[DataSaver] = None,
        loader_cls: Optional[DataLoader] = None,
    ) -> None:
        # != operator on boolean is XOR
        if bool(saver_cls is not None) != bool(loader_cls is not None):
            raise ValueError(
                "Must pass both `saver` and `loader` or neither. Currently received: "
                f"`saver`: `{saver_cls}`; `loader`: `{loader_cls}`"
            )

        if saver_cls is not None:
            # materialized_path
            materialized_path = self._materialized_path(data_version, saver_cls)
            saver = saver_cls(path=str(materialized_path.absolute()))
            loader = loader_cls(path=str(materialized_path.absolute()))
        else:
            saver = None
            loader = None

        self.path.mkdir(exist_ok=True)
        result_path = self._path_from_data_version(data_version)
        stored_result = StoredResult.new(value=result, saver=saver, loader=loader)
        self._write_result(result_path, stored_result)

    @override
    def get(self, data_version: str) -> Optional[Any]:
        result_path = self._path_from_data_version(data_version)
        stored_result = self._load_result_from_path(result_path)

        if stored_result is None:
            return None

        return stored_result.value

    @override
    def delete(self, data_version: str) -> None:
        result_path = self._path_from_data_version(data_version)
        result_path.unlink(missing_ok=True)

    @override
    def delete_all(self) -> None:
        shutil.rmtree(self.path)
        self.path.mkdir(exist_ok=True)

    def delete_expired(self) -> None:
        for file_path in self.path.iterdir():
            stored_result = self._load_result_from_path(file_path)
            if stored_result and stored_result.expired:
                file_path.unlink(missing_ok=True)



import logging
from typing import Any, Dict, List, Union

import diskcache

from hamilton import driver, graph_types, lifecycle, node

logger = logging.getLogger(__name__)


def _bytes_to_mb(kb: int) -> float:
    return kb / (1024**2)


def evict_all_except(nodes_to_keep: Dict[str, node.Node], cache: diskcache.Cache) -> int:
    """Evicts all nodes and node version except those passed.
    Remaining nodes may have multiple entries for different input values
    """
    nodes_history: Dict[str, List[str]] = cache.get(key=DiskCacheAdapter.nodes_history_key)  # type: ignore

    new_nodes_history = dict()
    eviction_counter = 0
    for node_name, history in nodes_history.items():
        if len(history) < 1:
            continue

        if node_name in nodes_to_keep.keys():
            node_to_keep = nodes_to_keep[node_name]
            hash_to_keep = graph_types.hash_source_code(node_to_keep.callable, strip=True)
            history.remove(hash_to_keep)
            new_nodes_history[node_name] = [hash_to_keep]

        for hash_to_evict in history:
            cache.evict(tag=f"{node_name}.{hash_to_evict}")
            eviction_counter += 1

    cache.set(key=DiskCacheAdapter.nodes_history_key, value=new_nodes_history)
    return eviction_counter


def evict_all_except_driver(dr: driver.Driver) -> dict:
    """Wrap the utility `evict_all_except` to receive a driver.Driver object"""
    cache_hooks = [
        adapter for adapter in dr.adapter.adapters if isinstance(adapter, DiskCacheAdapter)
    ]

    if len(cache_hooks) == 0:
        raise AssertionError("0 `h_diskcache.CacheHook` defined for this Driver")
    elif len(cache_hooks) > 1:
        raise AssertionError(">1 `h_diskcache.CacheHook` defined for this Driver")

    cache: diskcache.Cache = cache_hooks[0].cache
    volume_before = cache.volume()
    eviction_counter = evict_all_except(nodes_to_keep=dr.graph.nodes, cache=cache)
    volume_after = cache.volume()
    volume_difference = volume_before - volume_after

    logger.info(f"Evicted: {_bytes_to_mb(volume_difference):.2f} MB")
    logger.debug(f"Evicted {eviction_counter} entries")
    logger.debug(f"Cache size after: {_bytes_to_mb(volume_after):.2f} MB")

    return dict(
        evicted_size_mb=_bytes_to_mb(volume_difference),
        eviction_counter=eviction_counter,
        size_after=_bytes_to_mb(volume_after),
    )


MAX_KWARGS_REPR_LENGTH = 200


class DiskCacheAdapter(
    lifecycle.NodeExecutionHook,
    lifecycle.GraphExecutionHook,
    lifecycle.NodeExecutionMethod,
):
    nodes_history_key: str = "_nodes_history"

    def __init__(
        self, cache_vars: Union[List[str], None] = None, cache_path: str = ".", **cache_settings
    ):
        self.cache_vars = cache_vars if cache_vars else []
        self.cache_path = cache_path
        self.cache = diskcache.Cache(directory=cache_path, **cache_settings)
        self.nodes_history: Dict[str, List[str]] = self.cache.get(
            key=DiskCacheAdapter.nodes_history_key, default=dict()
        )  # type: ignore
        self.used_nodes_hash: Dict[str, str] = dict()

        logger.warning(
            "The `DiskCacheAdapter` is deprecated and will be removed in Hamilton 2.0. "
            "Consider enabling the core caching feature via `Builder.with_cache()`. "
            "This might not be 1-to-1 replacement, so please reach out if there are missing features. "
            "See https://hamilton.dagworks.io/en/latest/concepts/caching/ to learn more."
        )

    def run_before_graph_execution(self, *, graph: graph_types.HamiltonGraph, **kwargs):
        """Set cache_vars to all nodes if not specified"""
        if self.cache_vars == []:
            self.cache_vars = [n.name for n in graph.nodes]

    def run_to_execute_node(
        self, *, node_name: str, node_callable: Any, node_kwargs: Dict[str, Any], **kwargs
    ):
        """Create hash key then use cached value if exist"""
        if node_name not in self.cache_vars:
            return node_callable(**node_kwargs)

        node_hash = graph_types.hash_source_code(node_callable, strip=True)
        self.used_nodes_hash[node_name] = node_hash
        cache_key = (node_hash, *node_kwargs.values())

        from_cache = self.cache.get(key=cache_key, default=None)
        if from_cache is not None:
            if logger.isEnabledFor(logging.DEBUG):
                node_kwargs_string = repr(node_kwargs)
                if len(node_kwargs_string) > MAX_KWARGS_REPR_LENGTH:  # limit size of log
                    node_kwargs_string = node_kwargs_string[0:MAX_KWARGS_REPR_LENGTH] + "..."
                logger.debug(f"{node_name} {node_kwargs_string}: from cache")
            return from_cache

        if logger.isEnabledFor(logging.DEBUG):
            node_kwargs_string = repr(node_kwargs)
            if len(node_kwargs_string) > MAX_KWARGS_REPR_LENGTH:  # limit size of log
                node_kwargs_string = node_kwargs_string[0:MAX_KWARGS_REPR_LENGTH] + "..."
            logger.debug(f"{node_name} {node_kwargs_string}: executed")
        self.nodes_history[node_name] = self.nodes_history.get(node_name, []) + [node_hash]
        return node_callable(**node_kwargs)

    def run_after_node_execution(self, *, node_name: str, node_kwargs: dict, result: Any, **kwargs):
        if node_name not in self.cache_vars:
            return

        node_hash = self.used_nodes_hash[node_name]
        cache_key = (node_hash, *node_kwargs.values())
        cache_tag = f"{node_name}.{node_hash}"
        # only adds if key doesn't exist
        self.cache.add(key=cache_key, value=result, tag=cache_tag)

    def run_after_graph_execution(self, *args, **kwargs):
        self.cache.set(key=DiskCacheAdapter.nodes_history_key, value=self.nodes_history)
        logger.info(f"Cache size: {_bytes_to_mb(self.cache.volume()):.2f} MB")
        self.cache.close()

    def run_before_node_execution(self, *args, **kwargs):
        pass



from typing import Any, Dict, List, Optional, Sequence

try:
    from typing import override
except ImportError:
    override = lambda x: x  # noqa E731

from hamilton.caching.cache_key import decode_key

from .base import MetadataStore, ResultStore, StoredResult
from .file import FileResultStore
from .sqlite import SQLiteMetadataStore


class InMemoryMetadataStore(MetadataStore):
    def __init__(self) -> None:
        self._data_versions: Dict[str, str] = {}  # {cache_key: data_version}
        self._cache_keys_by_run: Dict[str, List[str]] = {}  # {run_id: [cache_key]}
        self._run_ids: List[str] = []

    @override
    def __len__(self) -> int:
        """Number of unique ``cache_key`` values."""
        return len(self._data_versions.keys())

    @override
    def exists(self, cache_key: str) -> bool:
        """Indicate if ``cache_key`` exists and it can retrieve a ``data_version``."""
        return cache_key in self._data_versions.keys()

    @override
    def initialize(self, run_id: str) -> None:
        """Set up and log the beginning of the run."""
        self._cache_keys_by_run[run_id] = []
        self._run_ids.append(run_id)

    @override
    def set(self, cache_key: str, data_version: str, run_id: str, **kwargs) -> Optional[Any]:
        """Set the ``data_version`` for ``cache_key`` and associate it with the ``run_id``."""
        self._data_versions[cache_key] = data_version
        self._cache_keys_by_run[run_id].append(cache_key)

    @override
    def get(self, cache_key: str) -> Optional[str]:
        """Retrieve the ``data_version`` for ``cache_key``."""
        return self._data_versions.get(cache_key, None)

    @override
    def delete(self, cache_key: str) -> None:
        """Delete the ``data_version`` for ``cache_key``."""
        del self._data_versions[cache_key]

    @override
    def delete_all(self) -> None:
        """Delete all stored metadata."""
        self._data_versions.clear()

    def persist_to(self, metadata_store: Optional[MetadataStore] = None) -> None:
        """Persist in-memory metadata using another MetadataStore implementation.

        :param metadata_store: MetadataStore implementation to use for persistence.
            If None, a SQLiteMetadataStore is created with the default path "./.hamilton_cache".

        .. code-block:: python

            from hamilton import driver
            from hamilton.caching.stores.sqlite import SQLiteMetadataStore
            from hamilton.caching.stores.memory import InMemoryMetadataStore
            import my_dataflow

            dr = (
              driver.Builder()
              .with_modules(my_dataflow)
              .with_cache(metadata_store=InMemoryMetadataStore())
              .build()
            )

            # execute the Driver several time. This will populate the in-memory metadata store
            dr.execute(...)

            # persist to disk in-memory metadata
            dr.cache.metadata_store.persist_to(SQLiteMetadataStore(path="./.hamilton_cache"))

        """
        if metadata_store is None:
            metadata_store = SQLiteMetadataStore(path="./.hamilton_cache")

        for run_id in self._run_ids:
            metadata_store.initialize(run_id)

        for run_id, cache_keys in self._cache_keys_by_run.items():
            for cache_key in cache_keys:
                data_version = self._data_versions[cache_key]
                metadata_store.set(
                    cache_key=cache_key,
                    data_version=data_version,
                    run_id=run_id,
                )

    @classmethod
    def load_from(cls, metadata_store: MetadataStore) -> "InMemoryMetadataStore":
        """Load in-memory metadata from another MetadataStore instance.

        :param metadata_store: MetadataStore instance to load from.
        :return: InMemoryMetadataStore copy of the ``metadata_store``.

        .. code-block:: python

            from hamilton import driver
            from hamilton.caching.stores.sqlite import SQLiteMetadataStore
            from hamilton.caching.stores.memory import InMemoryMetadataStore
            import my_dataflow

            sqlite_metadata_store = SQLiteMetadataStore(path="./.hamilton_cache")
            in_memory_metadata_store = InMemoryMetadataStore.load_from(sqlite_metadata_store)

            # create the Driver with the in-memory metadata store
            dr = (
              driver.Builder()
              .with_modules(my_dataflow)
              .with_cache(metadata_store=in_memory_metadata_store)
              .build()
            )

        """
        in_memory_metadata_store = InMemoryMetadataStore()

        for run_id in metadata_store.get_run_ids():
            in_memory_metadata_store.initialize(run_id)

            for node_metadata in metadata_store.get_run(run_id):
                in_memory_metadata_store.set(
                    cache_key=node_metadata["cache_key"],
                    data_version=node_metadata["data_version"],
                    run_id=run_id,
                )

        return in_memory_metadata_store

    @override
    def get_run_ids(self) -> List[str]:
        """Return a list of all ``run_id`` values stored."""
        return self._run_ids

    @override
    def get_run(self, run_id: str) -> List[Dict[str, str]]:
        """Return a list of node metadata associated with a run."""
        if self._cache_keys_by_run.get(run_id, None) is None:
            raise IndexError(f"Run ID not found: {run_id}")

        nodes_metadata = []
        for cache_key in self._cache_keys_by_run[run_id]:
            decoded_key = decode_key(cache_key)
            nodes_metadata.append(
                dict(
                    cache_key=cache_key,
                    data_version=self._data_versions[cache_key],
                    node_name=decoded_key["node_name"],
                    code_version=decoded_key["code_version"],
                    dependencies_data_versions=decoded_key["dependencies_data_versions"],
                )
            )

        return nodes_metadata


class InMemoryResultStore(ResultStore):
    def __init__(self, persist_on_exit: bool = False) -> None:
        self._results: Dict[str, StoredResult] = {}  # {data_version: result}

    @override
    def exists(self, data_version: str) -> bool:
        return data_version in self._results.keys()

    # TODO handle materialization
    @override
    def set(self, data_version: str, result: Any, **kwargs) -> None:
        self._results[data_version] = StoredResult.new(value=result)

    @override
    def get(self, data_version: str) -> Optional[Any]:
        stored_result = self._results.get(data_version, None)
        if stored_result is None:
            return None

        return stored_result.value

    @override
    def delete(self, data_version: str) -> None:
        del self._results[data_version]

    @override
    def delete_all(self) -> None:
        self._results.clear()

    def delete_expired(self) -> None:
        to_delete = [
            data_version
            for data_version, stored_result in self._results.items()
            if stored_result.expired
        ]

        # first collect keys then delete because you can delete from dictionary
        # as you iterate through it
        for data_version in to_delete:
            self.delete(data_version)

    def persist_to(self, result_store: Optional[ResultStore] = None) -> None:
        """Persist in-memory results using another ``ResultStore`` implementation.

        :param result_store: ResultStore implementation to use for persistence.
            If None, a FileResultStore is created with the default path "./.hamilton_cache".
        """
        if result_store is None:
            result_store = FileResultStore(path="./.hamilton_cache")

        for data_version, stored_result in self._results.items():
            result_store.set(data_version, stored_result.value)

    @classmethod
    def load_from(
        cls,
        result_store: ResultStore,
        metadata_store: Optional[MetadataStore] = None,
        data_versions: Optional[Sequence[str]] = None,
    ) -> "InMemoryResultStore":
        """Load in-memory results from another ResultStore instance.

        Since result stores do not store an index of their keys, you must provide a
        ``MetadataStore`` instance or a list of ``data_version`` for which results
        should be loaded in memory.

        :param result_store: ``ResultStore`` instance to load results from.
        :param metadata_store: ``MetadataStore`` instance from which all ``data_version`` are retrieved.
        :return: InMemoryResultStore copy of the ``result_store``.

        .. code-block:: python

            from hamilton import driver
            from hamilton.caching.stores.sqlite import SQLiteMetadataStore
            from hamilton.caching.stores.memory import InMemoryMetadataStore
            import my_dataflow

            sqlite_metadata_store = SQLiteMetadataStore(path="./.hamilton_cache")
            in_memory_metadata_store = InMemoryMetadataStore.load_from(sqlite_metadata_store)

            # create the Driver with the in-memory metadata store
            dr = (
              driver.Builder()
              .with_modules(my_dataflow)
              .with_cache(metadata_store=in_memory_metadata_store)
              .build()
            )


        """
        if metadata_store is None and data_versions is None:
            raise ValueError(
                "A `metadata_store` or `data_versions` must be provided to load results."
            )

        in_memory_result_store = InMemoryResultStore()

        data_versions_to_retrieve = set()
        if data_versions is not None:
            data_versions_to_retrieve.update(data_versions)

        if metadata_store is not None:
            for run_id in metadata_store.get_run_ids():
                for node_metadata in metadata_store.get_run(run_id):
                    data_versions_to_retrieve.add(node_metadata["data_version"])

        for data_version in data_versions_to_retrieve:
            # TODO disambiguate "result is None" from the sentinel value when `data_version`
            # is not found in `result_store`.
            result = result_store.get(data_version)
            in_memory_result_store.set(data_version, result)

        return in_memory_result_store



from hamilton.htypes import Collect, Parallelizable


def steps(number_of_steps: int) -> Parallelizable[int]:
    yield from range(number_of_steps)


def step_squared(steps: int, delay_seconds: float) -> int:
    import time

    time.sleep(delay_seconds)
    return steps**2


def step_cubed(steps: int) -> int:
    return steps**3


def step_squared_plus_step_cubed(step_squared: int, step_cubed: int) -> int:
    return step_squared + step_cubed


def sum_step_squared_plus_step_cubed(step_squared_plus_step_cubed: Collect[int]) -> int:
    out = 0
    for step in step_squared_plus_step_cubed:
        out += step
    return out


def final(sum_step_squared_plus_step_cubed: int) -> int:
    return sum_step_squared_plus_step_cubed



import dataclasses
from os import PathLike
from typing import Any, Collection, Dict, Tuple, Type, Union

try:
    import xgboost
except ImportError as e:
    raise NotImplementedError("XGBoost is not installed.") from e


from hamilton import registry
from hamilton.io import utils
from hamilton.io.data_adapters import DataLoader, DataSaver

XGBOOST_MODEL_TYPES = [xgboost.XGBModel, xgboost.Booster]
XGBOOST_MODEL_TYPES_ANNOTATION = Union[xgboost.XGBModel, xgboost.Booster]


@dataclasses.dataclass
class XGBoostJsonWriter(DataSaver):
    """Write XGBoost models and boosters to json format
    See differences with pickle format: https://xgboost.readthedocs.io/en/stable/tutorials/saving_model.html
    """

    path: Union[str, PathLike]

    @classmethod
    def applicable_types(cls) -> Collection[Type]:
        return XGBOOST_MODEL_TYPES

    def save_data(self, data: XGBOOST_MODEL_TYPES_ANNOTATION) -> Dict[str, Any]:
        data.save_model(self.path)
        return utils.get_file_metadata(self.path)

    @classmethod
    def name(cls) -> str:
        return "json"


@dataclasses.dataclass
class XGBoostJsonReader(DataLoader):
    """Load XGBoost models and boosters to json format
    See differences with pickle format: https://xgboost.readthedocs.io/en/stable/tutorials/saving_model.html
    """

    path: Union[str, bytearray, PathLike]

    @classmethod
    def applicable_types(cls) -> Collection[Type]:
        return XGBOOST_MODEL_TYPES

    def load_data(self, type_: Type) -> Tuple[XGBOOST_MODEL_TYPES_ANNOTATION, Dict[str, Any]]:
        model = type_()
        model.load_model(self.path)
        metadata = utils.get_file_metadata(self.path)
        return model, metadata

    @classmethod
    def name(cls) -> str:
        return "json"


def register_data_loaders():
    for loader in [
        XGBoostJsonReader,
        XGBoostJsonWriter,
    ]:
        registry.register_adapter(loader)


register_data_loaders()

COLUMN_FRIENDLY_DF_TYPE = False



import abc
import pickle
from datetime import datetime, timedelta, timezone
from typing import Any, Optional, Sequence, Tuple, Type

from hamilton.htypes import custom_subclass_check
from hamilton.io.data_adapters import DataLoader, DataSaver
from hamilton.registry import LOADER_REGISTRY, SAVER_REGISTRY


class ResultRetrievalError(Exception):
    """Raised by the SmartCacheAdapter when ResultStore.get() fails."""


# TODO Currently, this check is done when data needs to be saved.
# Ideally, it would be done earlier in the caching lifecycle.
def search_data_adapter_registry(
    name: str, type_: type
) -> Tuple[Type[DataSaver], Type[DataLoader]]:
    """Find pair of DataSaver and DataLoader registered with `name` and supporting `type_`"""
    if name not in SAVER_REGISTRY or name not in LOADER_REGISTRY:
        raise KeyError(
            f"{name} isn't associated to both a DataLoader and a DataSaver. "
            "Default saver/loader pairs include `json`, `file`, `pickle`, `parquet`, `csv`, "
            "`feather`, `orc`, `excel`. More pairs may be available through plugins."
        )

    try:
        saver_cls = next(
            saver_cls
            for saver_cls in SAVER_REGISTRY[name]
            if any(
                custom_subclass_check(type_, applicable_type)
                for applicable_type in saver_cls.applicable_types()
            )
        )
    except StopIteration as e:
        raise KeyError(f"{name} doesn't have any DataSaver supporting type {type_}") from e

    try:
        loader_cls = next(
            loader_cls
            for loader_cls in LOADER_REGISTRY[name]
            if any(
                custom_subclass_check(type_, applicable_type)
                for applicable_type in loader_cls.applicable_types()
            )
        )
    except StopIteration as e:
        raise KeyError(f"{name} doesn't have any DataLoader supporting type {type_}") from e

    return saver_cls, loader_cls


class ResultStore(abc.ABC):
    @abc.abstractmethod
    def set(self, data_version: str, result: Any, **kwargs) -> None:
        """Store ``result`` keyed by ``data_version``."""

    @abc.abstractmethod
    def get(self, data_version: str, **kwargs) -> Optional[Any]:
        """Try to retrieve ``result`` keyed by ``data_version``.
        If retrieval misses, return ``None``.
        """

    @abc.abstractmethod
    def delete(self, data_version: str) -> None:
        """Delete ``result`` keyed by ``data_version``."""

    @abc.abstractmethod
    def delete_all(self) -> None:
        """Delete all stored results."""

    @abc.abstractmethod
    def exists(self, data_version: str) -> bool:
        """boolean check if a ``result`` is found for ``data_version``
        If True, ``.get()`` should successfully retrieve the ``result``.
        """


class MetadataStore(abc.ABC):
    @abc.abstractmethod
    def __len__(self) -> int:
        """Return the number of cache_keys in the metadata store"""

    @abc.abstractmethod
    def initialize(self, run_id: str) -> None:
        """Setup the metadata store and log the start of the run"""

    @abc.abstractmethod
    def set(self, cache_key: str, data_version: str, **kwargs) -> Optional[Any]:
        """Store the mapping ``cache_key -> data_version``.
        Can include other metadata (e.g., node name, run id, code version) depending
        on the implementation.
        """

    @abc.abstractmethod
    def get(self, cache_key: str, **kwargs) -> Optional[str]:
        """Try to retrieve ``data_version`` keyed by ``cache_key``.
        If retrieval misses return ``None``.
        """

    @abc.abstractmethod
    def delete(self, cache_key: str) -> None:
        """Delete ``data_version`` keyed by ``cache_key``."""

    @abc.abstractmethod
    def delete_all(self) -> None:
        """Delete all stored metadata."""

    @abc.abstractmethod
    def exists(self, cache_key: str) -> bool:
        """boolean check if a ``data_version`` is found for ``cache_key``
        If True, ``.get()`` should successfully retrieve the ``data_version``.
        """

    @abc.abstractmethod
    def get_run_ids(self) -> Sequence[str]:
        """Return a list of run ids, sorted from oldest to newest start time.
        A ``run_id`` is registered when the metadata_store ``.initialize()`` is called.
        """

    @abc.abstractmethod
    def get_run(self, run_id: str) -> Sequence[dict]:
        """Return a list of node metadata associated with a run.

        For each node, the metadata should include ``cache_key`` (created or used)
        and ``data_version``. These values allow to manually query the MetadataStore
        or ResultStore.

        Decoding the ``cache_key`` gives the ``node_name``, ``code_version``, and
        ``dependencies_data_versions``. Individual implementations may add more
        information or decode the ``cache_key`` before returning metadata.
        """

    @property
    def size(self) -> int:
        """Number of unique entries (i.e., cache_keys) in the metadata_store"""
        return self.__len__()

    @property
    def last_run_id(self) -> str:
        """Return"""
        return self.get_run_ids()[-1]

    def get_last_run(self) -> Any:
        """Return the metadata from the last started run."""
        return self.get_run(self.last_run_id)


# TODO refactor the association between StoredResult, MetadataStore, and ResultStore
# to load data using the `DataLoader` class and kwargs instead of pickling the instantiated
# DataLoader object. This would be safer across Hamilton versions.
class StoredResult:
    def __init__(
        self,
        value: Any,
        expires_at=None,
        saver=None,
        loader=None,
    ):
        self.value = value
        self.expires_at = expires_at
        self.saver = saver
        self.loader = loader

    @classmethod
    def new(
        cls,
        value: Any,
        expires_in: Optional[timedelta] = None,
        saver: Optional[DataSaver] = None,
        loader: Optional[DataLoader] = None,
    ) -> "StoredResult":
        if expires_in is not None and not isinstance(expires_in, timedelta):
            expires_in = timedelta(seconds=expires_in)

        # != operator on boolean is XOR
        if bool(saver is not None) != bool(loader is not None):
            raise ValueError(
                "Must pass both `saver` and `loader` or neither. Currently received: "
                f"`saver`: `{saver}`; `loader`: `{loader}`"
            )

        return cls(
            value=value,
            expires_at=(datetime.now(tz=timezone.utc) + expires_in) if expires_in else None,
            saver=saver,
            loader=loader,
        )

    @property
    def expired(self) -> bool:
        return self.expires_at is not None and datetime.now(tz=timezone.utc) >= self.expires_at

    @property
    def expires_in(self) -> int:
        if self.expires_at:
            return int(self.expires_at.timestamp() - datetime.now(tz=timezone.utc).timestamp())

        return -1

    def save(self) -> bytes:
        """Receives pickleable data or DataLoader to use to load the real data"""
        if self.saver is not None:
            self.saver.save_data(data=self.value)
            to_pickle = self.loader
        else:
            to_pickle = self.value

        return pickle.dumps(to_pickle)

    @classmethod
    def load(cls, raw: bytes) -> "StoredResult":
        """Reads the raw bytes from disk and sets `StoredResult.data`"""
        loaded = pickle.loads(raw)
        if isinstance(loaded, DataLoader):
            loader = loaded
            result, metadata = loader.load_data(None)
        else:
            loader = None
            result = loaded

        return StoredResult.new(value=result)



import pytest

from hamilton import driver
from hamilton.execution.executors import SynchronousLocalTaskExecutor
from hamilton.lifecycle.default import GracefulErrorAdapter

from .resources import graceful_parallel


def _make_driver(
    test_state: str,
    try_parallel=True,
    fail_first=False,
    allow_injection=True,
):
    local_executor = SynchronousLocalTaskExecutor()
    dr = (
        driver.Builder()
        .enable_dynamic_execution(allow_experimental_mode=True)
        .with_modules(graceful_parallel)
        .with_remote_executor(local_executor)
        .with_adapters(
            GracefulErrorAdapter(
                error_to_catch=Exception,
                sentinel_value=None,
                try_all_parallel=try_parallel,
                allow_injection=allow_injection,
            )
        )
        .with_config(
            {
                "test_state": test_state,
                "test_front": "bad" if fail_first else "good",
            },
        )
        .build()
    )
    return dr


@pytest.mark.parametrize(
    "test_state, try_parallel, fail_first, expected_value",
    [
        ("early", True, False, [None]),
        ("middle", True, False, [0.0, 6.0, 12.0, 18.0, 24.0, None]),
        ("middle", False, False, [None]),
        ("pass", True, False, [0.0, 6.0, 12.0, 18.0, 24.0, 30.0, None, None, None, None]),
        ("pass", False, True, [None]),
    ],
    ids=[
        "test_early_parallel_fail",
        "test_middle_parallel_fail",
        "test_middle_parallel_as_single",
        "test_inside_block_errors",
        "test_pre_parallel_fail",
    ],
)
def test_parallel_graceful_simple(test_state, try_parallel, fail_first, expected_value) -> None:
    # Fail before the anything is yielded: should get 1 failure and skip all
    dr = _make_driver(
        test_state,
        try_parallel,
        fail_first,
    )
    ans = dr.execute(
        ["distro_gather"],
        inputs={"n": 10},
    )
    assert ans["distro_gather"] == expected_value


def test_parallel_gather_injection() -> None:
    dr = _make_driver(
        "pass",
        try_parallel=True,
        fail_first=False,
    )
    ans = dr.execute(
        ["distro_end"],
        inputs={"n": 10},
    )
    assert ans["distro_end"] == [
        [0.0, None],
        [6.0, None],
        [12.0, None],
        [18.0, None],
        [24.0, 13],
        [30.0, 16],
        [None, 19],
        [None, 22],
        [None, 25],
        [None, 28],
    ]

    # show that disallowing injection skips over the gathering step
    dr = _make_driver(
        "pass",
        try_parallel=True,
        fail_first=False,
        allow_injection=False,
    )
    ans = dr.execute(
        ["distro_end"],
        inputs={"n": 10},
    )
    assert ans["distro_end"] == [
        None,
        None,
        None,
        None,
        [24.0, 13],
        [30.0, 16],
        None,
        None,
        None,
        None,
    ]



# input
def number_of_steps() -> int:
    return 5


# process
def steps(number_of_steps: int) -> int:
    return number_of_steps


# process
def step_squared(steps: int) -> int:
    print("squaring step {}".format(steps))
    return steps**2


# process
def step_cubed(steps: int) -> int:
    print("cubing step {}".format(steps))
    return steps**3


def step_squared_plus_step_cubed(step_squared: int, step_cubed: int) -> int:
    print("adding step squared and step cubed")
    return step_squared + step_cubed


def sum_step_squared_plus_step_cubed(step_squared_plus_step_cubed: int) -> int:
    print("summing step squared")
    return step_squared_plus_step_cubed


# final
def final(sum_step_squared_plus_step_cubed: int) -> int:
    print("finalizing")
    return sum_step_squared_plus_step_cubed


def _calc():
    number_of_steps_ = number_of_steps()
    steps_ = steps(number_of_steps_)
    step_squared_ = step_squared(steps_)
    step_cubed_ = step_cubed(steps_)
    step_squared_plus_step_cubed_ = step_squared_plus_step_cubed(step_squared_, step_cubed_)
    sum_step_squared_plus_step_cubed_ = sum_step_squared_plus_step_cubed(
        step_squared_plus_step_cubed_
    )
    final_ = final(sum_step_squared_plus_step_cubed_)
    return final_



from typing import Any, Callable, Collection, Dict, List, Optional

from . import base

"""Decorators that handle the configuration of a function. These can be viewed as
replacing if/else/switch statements in standard dataflow definition libraries"""


class ConfigResolver:
    """Base class for resolving configuration so we can share the tooling between different functions."""

    def __init__(self, resolves: Callable[[Dict[str, Any]], bool], config_used: List[str]):
        self.resolves = resolves
        self._config_used = config_used

    @property
    def optional_config(self) -> Dict[str, Any]:
        """Gives the optional configuration for this resolver -- to be used by the @config decorator."""
        return {key: None for key in self._config_used}

    def __call__(self, config: Dict[str, Any]) -> bool:
        return self.resolves(config)

    @staticmethod
    def when(**key_value_pairs) -> "ConfigResolver":
        """Gives a resolver that resolves iff all keys in the config are equal to the corresponding value.


        :param key_value_pairs: Keys and corresponding values to look up in the config
        :return: a configuration decorator
        """

        def resolves(configuration: Dict[str, Any]) -> bool:
            return all(value == configuration.get(key) for key, value in key_value_pairs.items())

        return ConfigResolver(resolves, config_used=list(key_value_pairs.keys()))

    @staticmethod
    def when_not(**key_value_pairs: Any) -> "ConfigResolver":
        """Gives a resolver that resolves iff the keys in the config are all not equal to the corresponding value

        :param key_value_pairs: Keys and corresponding values to look up in the config
        :return: a configuration decorator
        """

        def resolves(configuration: Dict[str, Any]) -> bool:
            return all(value != configuration.get(key) for key, value in key_value_pairs.items())

        return ConfigResolver(resolves, config_used=list(key_value_pairs.keys()))

    @staticmethod
    def when_in(**key_value_group_pairs: Collection[Any]) -> "ConfigResolver":
        """Gives a resolver that the function if all the
        values corresponding to the config keys are equal to one of items in the list of values.

        :param key_value_group_pairs: pairs of key-value mappings where the value is a list of possible values
        :return: a configuration decorator
        """

        def resolves(configuration: Dict[str, Any]) -> bool:
            return all(
                configuration.get(key) in value for key, value in key_value_group_pairs.items()
            )

        return ConfigResolver(resolves, config_used=list(key_value_group_pairs.keys()))

    @staticmethod
    def when_not_in(**key_value_group_pairs: Collection[Any]) -> "ConfigResolver":
        """Gives a decorator that resolves the function only if none of the keys are in the list of values.

        :param key_value_group_pairs: pairs of key-value mappings where the value is a list of possible values
        :return: a configuration decorator
        """

        def resolves(configuration: Dict[str, Any]) -> bool:
            return all(
                configuration.get(key) not in value for key, value in key_value_group_pairs.items()
            )

        return ConfigResolver(resolves, config_used=list(key_value_group_pairs.keys()))


class config(base.NodeResolver):
    """Decorator class that determines whether a function should be in the DAG based on some configuration variable.

    Notes:

    1. Currently, functions that exist in all configurations have to be disjoint.

    2. There is currently no ``@config.otherwise(...)`` decorator, so make sure to have ``config.when`` specify set of \
    configuration possibilities. Any missing cases will not have that output (and subsequent downstream functions \
    may error out if they ask for it).

    3. To make this easier, we have a few more ``@config`` decorators:
        * ``@config.when_not(param=value)`` Will be included if the parameter is _not_ equal to the value specified.
        * ``@config.when_in(param=[value1, value2, ...])`` Will be included if the parameter is equal to one of the \
        specified values.
        * ``@config.when_not_in(param=[value1, value2, ...])`` Will be included if the parameter is not equal to any \
        of the specified values.
        * ``@config`` If you're feeling adventurous, you can pass in a lambda function that takes in the entire \
        configuration and resolves to ``True`` or ``False``. You probably don't want to do this.

    Example:

    .. code-block:: python

       @config.when_in(business_line=["mens","kids"], region=["uk"])
       def LEAD_LOG_BASS_MODEL_TIMES_TREND(
            TREND_BSTS_WOMENS_ACQUISITIONS: pd.Series,
            LEAD_LOG_BASS_MODEL_SIGNUPS_NON_REFERRAL: pd.Series) -> pd.Series:
            # logic
            ...

    Example - use of `__suffix` to differentiate between functions with the same name. This is required if you want to
    use the same function name in multiple configurations. Hamilton will automatically drop the suffix for you. The
    following will ensure only one function is registered with the name `my_transform`:

    .. code-block:: python

       @config.when(region="us")
       def my_transform__us(some_input: pd.Series, some_input_b: pd.Series) -> pd.Series:
            # logic
            ...

       @config.when(region="uk")
       def my_transform__uk(some_input: pd.Series, some_input_c: pd.Series) -> pd.Series:
            # logic
            ...


    ``@config`` If you're feeling adventurous, you can pass in a lambda function that takes in the entire configuration\
    and resolves to ``True`` or ``False``. You probably don't want to do this.
    """

    def __init__(
        self,
        resolves: Callable[[Dict[str, Any]], bool],
        target_name: str = None,
        config_used: List[str] = None,
    ):
        """Decorator that resolves a function based on the configuration...

        :param resolves: the python function to use to resolve whether the wrapped function should exist in the graph \
        or not.
        :param target_name: Optional. The name of the "function"/"node" that we want to attach @config to.
        :param config_used: Optional. The list of config names that this function uses.
        """
        self.does_resolve = resolves
        self.target_name = target_name
        self._config_used = config_used

    def required_config(self) -> Optional[List[str]]:
        """This returns the required configuration elements. Note that "none"
        is a sentinel value that means that we actaully don't know what
        it uses. If either required or optional configs are None, we
        pass the entire configuration.

        Note that this can still return None due to the @config(resolver) decorator.
        We will likely be deprecating this in 2.0, in favor of a (to be added) config.custom.
        Still thinking this over...

        :return: The list of required config elements, or None if we don't have any idea.
        """
        return None if self._config_used is None else []

    def optional_config(self) -> Optional[Dict[str, Any]]:
        """Everything is optional with None as the required value"""
        return {key: None for key in self._config_used} if self._config_used is not None else None

    def _get_function_name(self, fn: Callable) -> str:
        if self.target_name is not None:
            return self.target_name
        return base.sanitize_function_name(fn.__name__)

    def resolve(self, fn, config: Dict[str, Any]) -> Callable:
        if not self.does_resolve(config):
            return None
        # attaches config keys used to resolve function
        fn.__config_decorated__ = (
            self._config_used if self._config_used is not None else ["__UNKNOWN__"]
        )
        fn.__original_name__ = fn.__name__
        fn.__name__ = self._get_function_name(fn)  # TODO -- copy function to not mutate it
        return fn

    def validate(self, fn):
        if fn.__name__.endswith("__"):
            raise base.InvalidDecoratorException(
                "Config will always use the portion of the function name before the last __. For example, signups__v2 will map to signups, whereas"
            )

    @staticmethod
    def when(name=None, **key_value_pairs) -> "config":
        """Yields a decorator that resolves the function if all keys in the config are equal to the corresponding value.


        :param key_value_pairs: Keys and corresponding values to look up in the config
        :return: a configuration decorator
        """
        resolver = ConfigResolver.when(**key_value_pairs)
        return config(resolver, target_name=name, config_used=list(resolver.optional_config))

    @staticmethod
    def when_not(name=None, **key_value_pairs: Any) -> "config":
        """Yields a decorator that resolves the function if none keys in the config are equal to the corresponding value

        ``@config.when_not(param=value)`` will be included if the parameter is _not_ equal to the value specified.

        :param key_value_pairs: Keys and corresponding values to look up in the config
        :return: a configuration decorator
        """

        resolver = ConfigResolver.when_not(**key_value_pairs)
        return config(resolver, target_name=name, config_used=list(resolver.optional_config))

    @staticmethod
    def when_in(name=None, **key_value_group_pairs: Collection[Any]) -> "config":
        """Yields a decorator that resolves the function if all of the
        values corresponding to the config keys are equal to one of items in the list of values.

        ``@config.when_in(param=[value1, value2, ...])`` Will be included if the parameter is equal to one of the \
        specified values.

        :param key_value_group_pairs: pairs of key-value mappings where the value is a list of possible values
        :return: a configuration decorator
        """

        resolver = ConfigResolver.when_in(**key_value_group_pairs)
        return config(resolver, target_name=name, config_used=list(resolver.optional_config))

    @staticmethod
    def when_not_in(**key_value_group_pairs: Collection[Any]) -> "config":
        """Yields a decorator that resolves the function only if none of the keys are in the list of values.

        ``@config.when_not_in(param=[value1, value2, ...])`` Will be included if the parameter is not equal to any of \
        the specified values.

        :param key_value_group_pairs: pairs of key-value mappings where the value is a list of possible values
        :return: a configuration decorator

        .. code-block:: python

           @config.when_not_in(business_line=["mens","kids"], region=["uk"])
           def LEAD_LOG_BASS_MODEL_TIMES_TREND(
               TREND_BSTS_WOMENS_ACQUISITIONS: pd.Series,
               LEAD_LOG_BASS_MODEL_SIGNUPS_NON_REFERRAL: pd.Series) -> pd.Series:

        above will resolve for config has `{"business_line": "womens", "region": "us"}`,
        but not for configs that have `{"business_line": "mens", "region": "us"}`,
        `{"business_line": "kids", "region": "us"}`, or `{"region": "uk"}`.

        .. seealso::
            :ref:config.when_not
        """

        resolver = ConfigResolver.when_not_in(**key_value_group_pairs)
        return config(resolver, config_used=list(resolver.optional_config))


class hamilton_exclude(base.NodeResolver):
    """Decorator class that excludes a function from the DAG.

    The preferred way to hide functions from the Hamilton DAG is to prefix them with "_". However,
    for the exceptional case, it can be useful for decorating helper functions without the need to prefix
    them with "_" and use them either inside other nodes or in conjunction with ``step`` or ``apply_to``.

    .. code-block:: python

        @hamilton_exclude
        def helper(...) -> ...:
            '''This will not be part of the DAG'''
            ...

    You may also want to use this decorator for excluding functions in legacy code that would raise
    and error in Hamilton (for example missing type hints).
    """

    def __init__(self):
        pass

    def resolve(self, *args, **kwargs) -> Optional[Callable]:
        """Returning None defaults to not be included in the DAG.

        :param fn: Function to resolve
        :param config: DAG config
        :return: None to not be included in the DAG
        """
        return None

    def validate(self, fn):
        """Any function should work."""
        pass



import collections
import typing

import numpy as np
import pandas as pd
import pytest
from numpy import testing

from hamilton import base


def test_numpymatrixresult_int():
    """Tests the happy path of build_result of numpymatrixresult"""
    outputs = collections.OrderedDict(
        a=np.array([1, 7, 3, 7, 3, 6, 4, 9, 5, 0]), b=np.zeros(10), c=1
    )
    expected = np.array([[1, 7, 3, 7, 3, 6, 4, 9, 5, 0], np.zeros(10), np.ones(10)]).T
    actual = base.NumpyMatrixResult().build_result(**outputs)
    testing.assert_array_equal(actual, expected)


def test_numpymatrixresult_raise_length_mismatch():
    """Test raising an error build_result of numpymatrixresult"""
    outputs = collections.OrderedDict(
        a=np.array([1, 7, 3, 7, 3, 6, 4, 9, 5, 0]), b=np.array([1, 2, 3, 4, 5]), c=1
    )
    with pytest.raises(ValueError):
        base.NumpyMatrixResult().build_result(**outputs)


def test_SimplePythonGraphAdapter():
    """Tests that it delegates as intended"""

    class Foo(base.ResultMixin):
        @staticmethod
        def build_result(**outputs: typing.Dict[str, typing.Any]) -> typing.Any:
            outputs.update({"esoteric": "function"})
            return outputs

    spga = base.SimplePythonGraphAdapter(Foo())
    cols = {"a": "b"}
    expected = {"a": "b", "esoteric": "function"}
    actual = spga.build_result(**cols)
    assert actual == expected


def _gen_ints(n: int) -> typing.Iterator[int]:
    """Simple function to test that we can build results including generators."""
    yield from range(n)


class _Foo:
    """Dummy object used for testing."""

    def __init__(self, name: str):
        self.name = name

    def __eq__(self, other: typing.Any) -> bool:
        return isinstance(other, _Foo) and other.name == self.name


@pytest.mark.parametrize(
    "outputs,expected_result",
    [
        ({"a": 1}, pd.DataFrame([{"a": 1}])),
        ({"a": pd.Series([1, 2, 3])}, pd.DataFrame({"a": pd.Series([1, 2, 3])})),
        (
            {"a": pd.DataFrame({"a": [1, 2, 3], "b": [11, 12, 13]})},
            pd.DataFrame({"a": pd.Series([1, 2, 3]), "b": pd.Series([11, 12, 13])}),
        ),
        ({"a": [1, 2, 3]}, pd.DataFrame({"a": [1, 2, 3]})),
        ({"a": np.array([1, 2, 3])}, pd.DataFrame({"a": pd.Series([1, 2, 3])})),
        ({"a": {"b": 1, "c": "foo"}}, pd.DataFrame({"a": {"b": 1, "c": "foo"}})),
        ({"a": _gen_ints(3)}, pd.DataFrame({"a": pd.Series([0, 1, 2])})),
        ({"a": _Foo("bar")}, pd.DataFrame([{"a": _Foo("bar")}])),
        ({"a": 1, "bar": 2}, pd.DataFrame([{"a": 1, "bar": 2}])),
        (
            {"a": pd.Series([1, 2, 3]), "b": pd.Series([11, 12, 13])},
            pd.DataFrame({"a": pd.Series([1, 2, 3]), "b": pd.Series([11, 12, 13])}),
        ),
        (
            {"a": pd.Series([1, 2, 3]), "b": pd.Series([11, 12, 13]), "c": 1},
            pd.DataFrame(
                {"a": pd.Series([1, 2, 3]), "b": pd.Series([11, 12, 13]), "c": pd.Series([1, 1, 1])}
            ),
        ),
        (
            {
                "a": pd.Series([1, 2, 3]),
                "b": pd.Series([11, 12, 13]),
                "c": pd.Series([11, 12, 13]).index,
            },
            pd.DataFrame(
                {"a": pd.Series([1, 2, 3]), "b": pd.Series([11, 12, 13]), "c": pd.Series([0, 1, 2])}
            ),
        ),
        (
            {"a": [1, 2, 3], "b": [4, 5, 6]},
            pd.DataFrame({"a": pd.Series([1, 2, 3]), "b": pd.Series([4, 5, 6])}),
        ),
        (
            {"a": np.array([1, 2, 3]), "b": np.array([4, 5, 6])},
            pd.DataFrame({"a": pd.Series([1, 2, 3]), "b": pd.Series([4, 5, 6])}),
        ),
        (
            {"a": {"b": 1, "c": "foo"}, "d": {"b": 2}},
            pd.DataFrame({"a": pd.Series({"b": 1, "c": "foo"}), "d": pd.Series({"b": 2})}),
        ),
        (
            {"a": _gen_ints(3), "b": _gen_ints(3)},
            pd.DataFrame({"a": pd.Series([0, 1, 2]), "b": pd.Series([0, 1, 2])}),
        ),
        (
            {"a": pd.Series([1, 2, 3]), "b": 4},
            pd.DataFrame({"a": pd.Series([1, 2, 3]), "b": pd.Series([4, 4, 4])}),
        ),
        (
            {"a": [1, 2, 3], "b": 4},
            pd.DataFrame({"a": pd.Series([1, 2, 3]), "b": pd.Series([4, 4, 4])}),
        ),
        (
            {
                "a": {
                    "bar": 2,
                    "foo": 1,
                },
                "b": 4,
            },
            pd.DataFrame({"a": pd.Series([2, 1]), "b": pd.Series([4, 4])}).rename(
                index=lambda i: ["bar", "foo"][i]
            ),
        ),
        (
            {"a": pd.Series([1, 2, 3]), "b": pd.Series([4, 5, 6])},
            pd.DataFrame({"a": pd.Series([1, 2, 3]), "b": pd.Series([4, 5, 6])}),
        ),
        (
            {
                "a": pd.DataFrame({"a": [1, 2, 3], "b": [11, 12, 13]}),
                "b": pd.DataFrame({"c": [1, 3, 5], "d": [14, 15, 16]}),
            },
            pd.DataFrame(
                {"a.a": [1, 2, 3], "a.b": [11, 12, 13], "b.c": [1, 3, 5], "b.d": [14, 15, 16]}
            ),
        ),
        (
            {
                "a": pd.Series([1, 2, 3]),
                "b": pd.Series([11, 12, 13]),
                "c": pd.DataFrame({"d": [0, 0, 0]}),
            },
            pd.DataFrame(
                {"a": pd.Series([1, 2, 3]), "b": pd.Series([11, 12, 13]), "c.d": [0, 0, 0]}
            ),
        ),
    ],
    ids=[
        "test-single-scalar",
        "test-single-series",
        "test-single-dataframe",
        "test-single-list",
        "test-single-array",
        "test-single-dict",
        "test-single-generator",
        "test-single-object",
        "test-multiple-scalars",
        "test-multiple-series",
        "test-multiple-series-with-scalar",
        "test-multiple-series-with-index",
        "test-multiple-lists",
        "test-multiple-arrays",
        "test-multiple-dicts",
        "test-multiple-generators",
        "test-scalar-and-series",
        "test-scalar-and-list",
        "test-scalar-and-dict",
        "test-series-and-list",
        "test-multiple-dataframes",
        "test-multiple-series-with-dataframe",
    ],
)
def test_PandasDataFrameResult_build_result(outputs, expected_result):
    """Tests the happy case of PandasDataFrameResult.build_result()"""
    pdfr = base.PandasDataFrameResult()
    actual = pdfr.build_result(**outputs)
    pd.testing.assert_frame_equal(actual, expected_result)


@pytest.mark.parametrize(
    "outputs",
    [
        ({"a": [1, 2], "b": {"foo": "bar"}}),
        ({"a": [1, 2], "b": [3, 4, 5]}),
        ({"a": np.array([1, 2]), "b": np.array([3, 4, 5])}),
        ({"a": _gen_ints(3), "b": _gen_ints(4)}),
    ],
    ids=[
        "test-lists-and-dicts",
        "test-mismatched-lists",
        "test-mismatched-arrays",
        "test-mismatched-generators",
    ],
)
def test_PandasDataFrameResult_build_result_errors(outputs):
    """Tests the error case of PandasDataFrameResult.build_result()"""
    pdfr = base.PandasDataFrameResult()
    with pytest.raises(ValueError):
        pdfr.build_result(**outputs)


@pytest.mark.parametrize(
    "outputs,expected_result",
    [
        (
            {
                "a": pd.DataFrame({"a": [1, 2, 3], "z": [0, 0, 0]}),
                "b": pd.Series([4, 5, 6]),
                "c": 7,
                "d": [8, 9, 10],
            },
            pd.DataFrame(
                {
                    "a.a": [1, 2, 3],
                    "a.z": [0, 0, 0],
                    "b": [4, 5, 6],
                    "c": [7, 7, 7],
                    "d": [8, 9, 10],
                }
            ),
        ),
        (
            {
                "a": pd.DataFrame({"a": [1, 2, 3], "b": [11, 12, 13]}, index=[0, 1, 2]),
                "b": pd.DataFrame({"c": [1, 3, 5], "d": [14, 15, 16]}, index=[3, 4, 5]),
            },
            pd.DataFrame(
                {
                    "a.a": [1, 2, 3, None, None, None],
                    "a.b": [11, 12, 13, None, None, None],
                    "b.c": [None, None, None, 1, 3, 5],
                    "b.d": [None, None, None, 14, 15, 16],
                },
                index=[0, 1, 2, 3, 4, 5],
            ),
        ),
        (
            {
                "a": pd.Series([1, 2, 3], index=[1, 2, 3]),
                "c": pd.DataFrame({"d": [0, 0, 0], "e": [1, 1, 1]}),
                "b": pd.Series([11, 12, 13]),
                "f": pd.DataFrame({"g": [2, 2, 2], "h": [3, 3, 3]}, index=[1, 2, 3]),
            },
            pd.DataFrame(
                {
                    "a": [None, 1, 2, 3],
                    "c.d": [0, 0, 0, None],
                    "c.e": [1, 1, 1, None],
                    "b": [11, 12, 13, None],
                    "f.g": [None, 2, 2, 2],
                    "f.h": [None, 3, 3, 3],
                },
                index=[0, 1, 2, 3],
            ),
        ),
    ],
    ids=[
        "test-dataframe-scalar-series-list",
        "test-two-dataframes",
        "test-order-and-outer-join-preserved",
    ],
)
def test_PandasDataFrameResult_build_dataframe_with_dataframes(outputs, expected_result):
    """Tests build_dataframe_with_dataframes errors as expected"""
    pdfr = base.PandasDataFrameResult()
    actual = pdfr.build_dataframe_with_dataframes(outputs)
    pd.testing.assert_frame_equal(actual, expected_result)


# Still supporting old pandas version, although we should phase off...
int_64_index = "Index:::int64" if pd.__version__ >= "2.0.0" else "RangeIndex:::int64"

PD_VERSION = tuple(int(item) for item in pd.__version__.split("."))


@pytest.mark.parametrize(
    "outputs,expected_result",
    [
        ({"a": pd.Series([1, 2, 3])}, ({"RangeIndex:::int64": ["a"]}, {}, {})),
        (
            {"a": pd.Series([1, 2, 3]), "b": pd.Series([3, 4, 5])},
            ({"RangeIndex:::int64": ["a", "b"]}, {}, {}),
        ),
        (
            {
                "b": pd.Series(
                    [3, 4, 5], index=pd.DatetimeIndex(["2022-01", "2022-02", "2022-03"], freq="MS")
                )
            },
            (
                {"DatetimeIndex:::datetime64[ns]": ["b"]},
                {"DatetimeIndex:::datetime64[ns]": ["b"]},
                {},
            ),
        ),
        ({"c": 1}, ({"no-index": ["c"]}, {}, {"no-index": ["c"]})),
        (
            {
                "a": pd.Series([1, 2, 3]),
                "b": 1,
                "c": pd.Series(
                    [3, 4, 5], index=pd.DatetimeIndex(["2022-01", "2022-02", "2022-03"], freq="MS")
                ),
            },
            (
                {
                    "DatetimeIndex:::datetime64[ns]": ["c"],
                    "RangeIndex:::int64": ["a"],
                    "no-index": ["b"],
                },
                {"DatetimeIndex:::datetime64[ns]": ["c"]},
                {"no-index": ["b"]},
            ),
        ),
        ({"a": pd.DataFrame({"a": [1, 2, 3]})}, ({"RangeIndex:::int64": ["a"]}, {}, {})),
        pytest.param(
            {"a": pd.Series([1, 2, 3]).index},
            ({"Index:::int64": ["a"]}, {}, {}),
            marks=pytest.mark.skipif(
                PD_VERSION < (2, 0, 0),
                reason="Pandas 2.0 changed default indices but we still " "support pandas <2.0",
            ),
        ),
        pytest.param(
            {"a": pd.Series([1, 2, 3]).index},
            ({"Int64Index:::int64": ["a"]}, {}, {}),
            marks=pytest.mark.skipif(
                PD_VERSION >= (2, 0, 0),
                reason="Pandas 2.0 changed default indices but we still " "support pandas <2.0",
            ),
        ),
    ],
    ids=[
        "int-index",
        "int-index-double",
        "ts-index",
        "no-index",
        "multiple-different-indexes",
        "df-index",
        "index-object-3-7",
        "index-object-3-8-plus",
    ],
)
def test_PandasDataFrameResult_pandas_index_types(outputs, expected_result):
    """Tests exercising the function to return pandas index types from outputs"""
    pdfr = base.PandasDataFrameResult()
    actual = pdfr.pandas_index_types(outputs)
    assert dict(actual[0]) == expected_result[0]
    assert dict(actual[1]) == expected_result[1]
    assert dict(actual[2]) == expected_result[2]


@pytest.mark.parametrize(
    "all_index_types,time_indexes,no_indexes,expected_result",
    [
        ({"foo": ["a", "b", "c"]}, {}, {}, True),
        ({"int-index": ["a"], "no-index": ["b"]}, {}, {"no-index": ["b"]}, True),
        ({"ts-1": ["a"], "ts-2": ["b"]}, {"ts-1": ["a"], "ts-2": ["b"]}, {}, False),
        ({"float-index": ["a"], "int-index": ["b"]}, {}, {}, False),
        ({"no-index": ["a", "b"]}, {}, {"no-index": ["a", "b"]}, False),
    ],
    ids=[
        "all-the-same",  # True
        "single-index-with-no-index",  # True
        "multiple-ts",  # False
        "multiple-indexes-not-ts",  # False
        "no-indexes-at-all",  # False4
    ],
)
def test_PandasDataFrameResult_check_pandas_index_types_match(
    all_index_types, time_indexes, no_indexes, expected_result
):
    """Tests exercising the function to determine whether pandas index types match"""
    # setup to test conditional if statement on logger level
    import logging

    logger = logging.getLogger("hamilton.base")  # get logger of base module.
    logger.setLevel(logging.DEBUG)
    pdfr = base.PandasDataFrameResult()
    actual = pdfr.check_pandas_index_types_match(all_index_types, time_indexes, no_indexes)
    assert actual == expected_result


@pytest.mark.parametrize(
    "outputs,expected_result",
    [
        ({"a": pd.Series([1, 2, 3])}, pd.DataFrame({"a": pd.Series([1, 2, 3])})),
        (
            {
                "a": pd.Series(
                    [1, 2, 3], index=pd.DatetimeIndex(["2022-01", "2022-02", "2022-03"], freq="MS")
                ),
                "b": pd.Series(
                    [3, 4, 5], index=pd.DatetimeIndex(["2022-01", "2022-02", "2022-03"], freq="MS")
                ),
            },
            pd.DataFrame(
                {
                    "a": pd.Series(
                        [1, 2, 3],
                        index=pd.DatetimeIndex(["2022-01", "2022-02", "2022-03"], freq="MS"),
                    ),
                    "b": pd.Series(
                        [3, 4, 5],
                        index=pd.DatetimeIndex(["2022-01", "2022-02", "2022-03"], freq="MS"),
                    ),
                }
            ),
        ),
        (
            {
                "a": pd.Series(
                    [1, 2, 3], index=pd.DatetimeIndex(["2022-01", "2022-02", "2022-03"], freq="MS")
                ),
                "b": 4,
            },
            pd.DataFrame(
                {
                    "a": pd.Series(
                        [1, 2, 3],
                        index=pd.DatetimeIndex(["2022-01", "2022-02", "2022-03"], freq="MS"),
                    ),
                    "b": 4,
                }
            ),
        ),
    ],
    ids=[
        "test-same-index-simple",
        "test-same-index-ts",
        "test-index-with-scalar",
    ],
)
def test_StrictIndexTypePandasDataFrameResult_build_result(outputs, expected_result):
    """Tests the happy case of StrictIndexTypePandasDataFrameResult.build_result()"""
    sitpdfr = base.StrictIndexTypePandasDataFrameResult()
    actual = sitpdfr.build_result(**outputs)
    pd.testing.assert_frame_equal(actual, expected_result)


@pytest.mark.parametrize(
    "outputs",
    [
        (
            {
                "a": pd.Series([1, 2, 3], index=[0, 1, 2]),
                "b": pd.Series([1, 2, 3], index=[0.0, 1.0, 2.0]),
            }
        ),
        (
            {
                "series1": pd.Series(
                    [1, 2, 3], index=pd.DatetimeIndex(["2022-01", "2022-02", "2022-03"], freq="MS")
                ),
                "series2": pd.Series(
                    [4, 5, 6],
                    index=pd.PeriodIndex(year=[2022, 2022, 2022], month=[1, 2, 3], freq="M"),
                ),
                "series3": pd.Series(
                    [4, 5, 6],
                    index=pd.PeriodIndex(
                        year=[2022, 2022, 2022], month=[1, 1, 1], day=[3, 4, 5], freq="B"
                    ),
                ),
                "series4": pd.Series(
                    [4, 5, 6],
                    index=pd.PeriodIndex(
                        year=[2022, 2022, 2022], month=[1, 1, 1], day=[4, 11, 18], freq="W"
                    ),
                ),
            }
        ),
    ],
    ids=[
        "test-int-float",
        "test-different-ts-indexes",
    ],
)
def test_StrictIndexTypePandasDataFrameResult_build_result_errors(outputs):
    """Tests the error case of StrictIndexTypePandasDataFrameResult.build_result()"""
    sitpdfr = base.StrictIndexTypePandasDataFrameResult()
    with pytest.raises(ValueError):
        sitpdfr.build_result(**outputs)



import json

from hamilton import driver, graph_types

from tests.resources.dynamic_parallelism import no_parallel

# `version` value isn't hardcoded because it's Python version-dependent
EXPECTED_JSON = {
    "nodes": [
        {
            "name": "final",
            "tags": {"module": "tests.resources.dynamic_parallelism.no_parallel"},
            "output_type": "int",
            "required_dependencies": ["sum_step_squared_plus_step_cubed"],
            "optional_dependencies": [],
            "source": 'def final(sum_step_squared_plus_step_cubed: int) -> int:\n    print("finalizing")\n    return sum_step_squared_plus_step_cubed\n',
            "documentation": "",
            "version": "27af6c2870dd4ca92335df4a9ff6c84da829d792a4c9f690f2f57193dbf3395a",
        },
        {
            "name": "number_of_steps",
            "tags": {"module": "tests.resources.dynamic_parallelism.no_parallel"},
            "output_type": "int",
            "required_dependencies": [],
            "optional_dependencies": [],
            "source": "def number_of_steps() -> int:\n    return 5\n",
            "documentation": "",
            "version": "48945782868c44bca1b159682d48bee3a1c1e9be73cf059c5087d0d3258fdc78",
        },
        {
            "name": "step_cubed",
            "tags": {"module": "tests.resources.dynamic_parallelism.no_parallel"},
            "output_type": "int",
            "required_dependencies": ["steps"],
            "optional_dependencies": [],
            "source": 'def step_cubed(steps: int) -> int:\n    print("cubing step {}".format(steps))\n    return steps**3\n',
            "documentation": "",
            "version": "f82b894f3d8a8de9cf798a41e08c1f38c8396e86c569cb73b7d78b4420f9dd99",
        },
        {
            "name": "step_squared",
            "tags": {"module": "tests.resources.dynamic_parallelism.no_parallel"},
            "output_type": "int",
            "required_dependencies": ["steps"],
            "optional_dependencies": [],
            "source": 'def step_squared(steps: int) -> int:\n    print("squaring step {}".format(steps))\n    return steps**2\n',
            "documentation": "",
            "version": "371bff5e02e091314e0ef2c3ca3003a496a119f9164d3e1827fe64d5f8c4aaba",
        },
        {
            "name": "step_squared_plus_step_cubed",
            "tags": {"module": "tests.resources.dynamic_parallelism.no_parallel"},
            "output_type": "int",
            "required_dependencies": ["step_cubed", "step_squared"],
            "optional_dependencies": [],
            "source": 'def step_squared_plus_step_cubed(step_squared: int, step_cubed: int) -> int:\n    print("adding step squared and step cubed")\n    return step_squared + step_cubed\n',
            "documentation": "",
            "version": "fe76eaaf43a6b79a706574ea6a8f00fe5b6439e3b8d7ceb879ab938b7722a1b3",
        },
        {
            "name": "steps",
            "tags": {"module": "tests.resources.dynamic_parallelism.no_parallel"},
            "output_type": "int",
            "required_dependencies": ["number_of_steps"],
            "optional_dependencies": [],
            "source": "def steps(number_of_steps: int) -> int:\n    return number_of_steps\n",
            "documentation": "",
            "version": "111ccfe2b5e8754e5ada1dae3ffc49d27b450a98e3be61f9308144aa5a444921",
        },
        {
            "name": "sum_step_squared_plus_step_cubed",
            "tags": {"module": "tests.resources.dynamic_parallelism.no_parallel"},
            "output_type": "int",
            "required_dependencies": ["step_squared_plus_step_cubed"],
            "optional_dependencies": [],
            "source": 'def sum_step_squared_plus_step_cubed(step_squared_plus_step_cubed: int) -> int:\n    print("summing step squared")\n    return step_squared_plus_step_cubed\n',
            "documentation": "",
            "version": "7831c9ed5f45e0b8ce3de34c3a5e2b4d4467b1ebc44cb55ea6ecae8f7dbf749d",
        },
    ]
}


def test_export_execution():
    dr = driver.Builder().with_modules(no_parallel).build()
    graph = graph_types.HamiltonGraph.from_graph(dr.graph)

    # can't hard code HamiltonNode.version because it depends on Python version
    for n in graph.nodes:
        for json_n in EXPECTED_JSON["nodes"]:
            if json_n["name"] == n.name:
                json_n["version"] = n.version

    json_str = dr.export_execution(["final"])
    assert json_str == json.dumps(EXPECTED_JSON)
